"""coBib's debugging config."""

from pathlib import Path

from cobib.config import config

config.logging.version = None

root = Path(__file__).parent
config.database.file = str((root / "example_literature.yaml").resolve())

config.utils.file_downloader.url_map[
    r"(.+)://quantum-journal.org/papers/([^/]+)"
] = r"\1://quantum-journal.org/papers/\2/pdf/"

config.utils.journal_abbreviations = [
    ("Annalen der Physik", "Ann. Phys."),
]



"""coBib's arXiv parser.

This parser is capable of generating `cobib.database.Entry` instances from a given arXiv ID.
It gathers the BibTex-encoded data from the arXiv API and parses the raw XML data.

Since v3.2.0 coBib will also automatically download the PDF version of the new entry. You can
configure the default download location via
`cobib.config.config.FileDownloaderConfig.default_location`.

Since v3.3.0 this parser even supports URLs from which an arXiv ID can be extracted directly.

The parser is registered under the `-a` and `--arxiv` command-line arguments of the
`cobib.commands.add.AddCommand`.

The following documentation is mostly inherited from the abstract interface
`cobib.parsers.base_parser`.
"""

import logging
import re
from collections import OrderedDict
from typing import Any, Dict

import requests
from bs4 import BeautifulSoup
from typing_extensions import override

from cobib.config import Event
from cobib.database import Entry

from .base_parser import Parser

LOGGER = logging.getLogger(__name__)

ARXIV_URL = "https://export.arxiv.org/api/query?id_list="
"""arXiv exporting URL taken from [here](https://arxiv.org/help/oa)."""
ARXIV_REGEX = r"(\d{4}.\d{4,5}|[a-z\-]+(\.[A-Z]{2})?\/\d{7})(v\d+)?"
"""A regex pattern used to match valid DOIs."""


class ArxivParser(Parser):
    """The arXiv Parser."""

    name = "arxiv"

    @override
    def parse(self, string: str) -> Dict[str, Entry]:
        string = Event.PreArxivParse.fire(string) or string

        try:
            match = re.search(ARXIV_REGEX, string)
            if match is None:
                raise AssertionError
        except AssertionError:
            msg = f"'{string}' is not a valid arXiv ID."
            LOGGER.warning(msg)
            return OrderedDict()
        arxiv_id = match.group(1)
        LOGGER.info("Gathering BibTex data for arXiv ID: %s.", arxiv_id)
        try:
            page = requests.get(ARXIV_URL + arxiv_id, timeout=10)
        except requests.exceptions.RequestException as err:
            LOGGER.error("An Exception occurred while trying to query the arXiv ID: %s.", arxiv_id)
            LOGGER.error(err)
            return OrderedDict()
        xml = BeautifulSoup(page.text, features="xml")
        if xml.feed.entry.title.contents[0] == "Error":
            msg = (
                "The arXiv API returned the following error: " + xml.feed.entry.summary.contents[0]
            )
            LOGGER.warning(msg)
            return OrderedDict()
        label = ""
        entry: Dict[str, Any] = {}
        entry["archivePrefix"] = "arXiv"
        for key in xml.feed.entry.findChildren(recursive=False):
            if "doi" in key.name:
                entry["doi"] = str(key.contents[0])
            elif key.name == "id":
                entry["arxivid"] = str(key.contents[0]).replace("http://arxiv.org/abs/", "")
                entry["eprint"] = str(key.contents[0])
            elif key.name == "primary_category":
                entry["primaryClass"] = str(key.attrs["term"])
            elif key.name == "published":
                # The year must also be stored as a string for compatibility reasons with
                # bibtexparser. However, we perform a conversion to an integer first, to ensure that
                # the year can actually be represented as such.
                entry["year"] = int(key.contents[0].split("-")[0])
                label += str(entry["year"])
            elif key.name == "title":
                entry["title"] = re.sub(r"\s+", " ", key.contents[0].strip().replace("\n", " "))
            elif key.name == "author":
                if "author" not in entry:
                    first = True
                    entry["author"] = ""
                name = [n.contents[0] for n in key.findChildren()][0]
                if first:
                    label = name.split()[-1] + label
                    first = False
                entry["author"] += f"{name} and "
            elif key.name == "summary":
                entry["abstract"] = re.sub(r"\s+", " ", key.contents[0].strip().replace("\n", " "))
            elif key.name == "link":
                if key.attrs.get("title", None) == "doi":
                    entry["url"] = key.attrs["href"]
                elif key.attrs.get("title", None) == "pdf":
                    entry["_download"] = key.attrs.get("href", "")
            else:
                LOGGER.warning("The key '%s' of this arXiv entry is not being processed!", key.name)
        if "doi" in entry:
            entry["ENTRYTYPE"] = "article"
        else:
            entry["ENTRYTYPE"] = "unpublished"
        # strip last 'and' from author field
        entry["author"] = entry["author"][:-5]
        bib = OrderedDict()
        bib[label] = Entry(label, entry)

        Event.PostArxivParse.fire(bib)

        return bib

    def dump(self, entry: Entry) -> None:
        """We cannot dump a generic entry as an arXiv ID."""
        LOGGER.error("Cannot dump an entry as an arXiv ID.")



"""Tests for coBib's file downloader utility."""

import tempfile
from os import remove
from typing import Any, Dict, Optional, Tuple

import pytest
import requests

from cobib.config import Event, config
from cobib.utils.file_downloader import FileDownloader
from cobib.utils.rel_path import RelPath

from .. import get_resource


@pytest.fixture
def setup_remove_content_length(monkeypatch: pytest.MonkeyPatch, enable: bool = True) -> None:
    """Setup method to remove the `content-length` from the response header.

    Args:
        monkeypatch: the built-in pytest fixture.
        enable: whether to enable this fixture.
    """
    if not enable:
        return

    def remove_content_length(*args, **kwargs):  # type: ignore
        """Mock function to remove `content-length` from response."""
        response = requests.request("get", *args, timeout=kwargs.pop("timeout", 30), **kwargs)
        response.headers.pop("content-length")
        return response

    monkeypatch.setattr(requests, "get", remove_content_length)


def test_downloader_singleton() -> None:
    """Test the FileDownloader is a Singleton."""
    f_d = FileDownloader()
    f_d2 = FileDownloader()
    assert f_d is f_d2


@pytest.mark.asyncio
async def test_download(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test the FileDownloader.download method.

    Args:
        monkeypatch: the built-in pytest fixture.
    """
    with tempfile.TemporaryDirectory() as tmpdirname:
        try:
            # ensure file does not exist yet
            remove(tmpdirname + "/dummy.pdf")
        except FileNotFoundError:
            pass
        # disable the PDF assertion method
        monkeypatch.setattr(FileDownloader, "_assert_pdf", lambda _: True)
        path = await FileDownloader().download(
            "https://gitlab.com/cobib/cobib/-/raw/master/tests/utils/__init__.py",
            "dummy",
            tmpdirname,
        )
        if path is None:
            pytest.skip("Likely, a requests error occured.")
        with open(get_resource("__init__.py", "utils"), "r", encoding="utf-8") as expected:
            with open(tmpdirname + "/dummy.pdf", "r", encoding="utf-8") as truth:
                assert expected.read() == truth.read()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ["setup_remove_content_length"],
    [
        [{"enable": False}],
        [{"enable": True}],
    ],
    indirect=["setup_remove_content_length"],
)
# pylint: disable=unused-argument,redefined-outer-name
async def test_skip_download_if_no_pdf(setup_remove_content_length: Any) -> None:
    """Test that download is skipped when file is not a PDF.

    Args:
        setup_remove_content_length: a custom fixture.
    """
    with tempfile.TemporaryDirectory() as tmpdirname:
        assert (
            await FileDownloader().download(
                "https://gitlab.com/cobib/cobib/-/raw/master/tests/utils/__init__.py",
                "dummy",
                tmpdirname,
            )
            is None
        )


@pytest.mark.asyncio
@pytest.mark.parametrize("overwrite", [False, True])
async def test_skip_download_if_exists(caplog: pytest.LogCaptureFixture, overwrite: bool) -> None:
    """Test that download is skipped when file already exists.

    Args:
        caplog: the built-in pytest fixture.
        overwrite: whether or not to overwrite the existing file.
    """
    with tempfile.TemporaryDirectory() as tmpdirname:
        open(  # pylint: disable=consider-using-with
            tmpdirname + "/dummy.pdf", "w", encoding="utf-8"
        ).close()
        await FileDownloader().download(
            "https://gitlab.com/cobib/cobib/-/raw/master/tests/utils/__init__.py",
            "dummy",
            folder=tmpdirname,
            overwrite=overwrite,
        )
        for mod, lvl, msg in caplog.record_tuples:
            if (
                mod == "cobib.utils.file_downloader"
                and lvl == 30
                and "already exists! Using that rather than downloading." in msg
            ):
                break
        else:
            assert overwrite, "This statement should only be reached when overwrite is enabled!"
            return
        assert not overwrite, "This statement should only be reached when overwrite is disabled!"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    ["setup_remove_content_length"],
    [
        [{"enable": False}],
        [{"enable": True}],
    ],
    indirect=["setup_remove_content_length"],
)
# pylint: disable=unused-argument,redefined-outer-name
async def test_download_with_url_map(setup_remove_content_length: Any) -> None:
    """Test the `config.utils.file_downloader.url_map` usage.

    We use a Quantum Journal article because they are open-source and, thus, do not require a Proxy
    to get access to.

    Args:
        setup_remove_content_length: a custom fixture.
    """
    try:
        config.load(get_resource("debug.py"))
        with tempfile.TemporaryDirectory() as tmpdirname:
            try:
                # ensure file does not exist yet
                remove(tmpdirname + "/dummy.pdf")
            except FileNotFoundError:
                pass
            path = await FileDownloader().download(
                "https://quantum-journal.org/papers/q-2021-06-17-479/",
                "dummy",
                tmpdirname,
            )
            if path is None:
                pytest.skip("Likely, a requests error occured.")
            assert path.path.exists()
            with open(path.path, "rb") as file:
                assert file.read().startswith(bytes("%PDF", "utf-8"))
    finally:
        config.defaults()


@pytest.mark.asyncio
async def test_gracefully_fail_download(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test gracefully failing downloads.

    Args:
        monkeypatch: the built-in pytest fixture.
    """

    def raise_exception(*args, **kwargs):  # type: ignore
        """Mock function to raise an Exception."""
        raise requests.exceptions.RequestException()

    monkeypatch.setattr(requests, "get", raise_exception)

    with tempfile.TemporaryDirectory() as tmpdirname:
        assert (
            await FileDownloader().download(
                "https://quantum-journal.org/papers/q-2021-06-17-479/",
                "dummy",
                tmpdirname,
            )
            is None
        )


@pytest.mark.asyncio
async def test_event_pre_download(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test the PreFileDownload event.

    Args:
        monkeypatch: the built-in pytest fixture.
    """

    @Event.PreFileDownload.subscribe
    def hook(
        url: str, label: str, folder: Optional[str], headers: Optional[Dict[str, str]]
    ) -> Optional[Tuple[str, str, Optional[str], Optional[Dict[str, str]]]]:
        label = "test"
        return (url, label, folder, headers)

    assert Event.PreFileDownload.validate()

    with tempfile.TemporaryDirectory() as tmpdirname:
        try:
            # ensure file does not exist yet
            remove(tmpdirname + "/test.pdf")
        except FileNotFoundError:
            pass
        # disable the PDF assertion method
        monkeypatch.setattr(FileDownloader, "_assert_pdf", lambda _: True)
        path = await FileDownloader().download(
            "https://gitlab.com/cobib/cobib/-/raw/master/tests/utils/__init__.py",
            "dummy",
            tmpdirname,
        )
        if path is None:
            pytest.skip("Likely, a requests error occured.")
        with open(get_resource("__init__.py", "utils"), "r", encoding="utf-8") as expected:
            with open(tmpdirname + "/test.pdf", "r", encoding="utf-8") as truth:
                assert expected.read() == truth.read()


@pytest.mark.asyncio
async def test_event_post_download(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test the PostFileDownload event.

    Args:
        monkeypatch: the built-in pytest fixture.
    """

    @Event.PostFileDownload.subscribe
    def hook(path: RelPath) -> Optional[RelPath]:  # type: ignore[return]
        path.path.rename(path.path.parent / "test.pdf")

    assert Event.PostFileDownload.validate()

    with tempfile.TemporaryDirectory() as tmpdirname:
        try:
            # ensure file does not exist yet
            remove(tmpdirname + "/test.pdf")
        except FileNotFoundError:
            pass
        # disable the PDF assertion method
        monkeypatch.setattr(FileDownloader, "_assert_pdf", lambda _: True)
        path = await FileDownloader().download(
            "https://gitlab.com/cobib/cobib/-/raw/master/tests/utils/__init__.py",
            "dummy",
            tmpdirname,
        )
        if path is None:
            pytest.skip("Likely, a requests error occured.")
        with open(get_resource("__init__.py", "utils"), "r", encoding="utf-8") as expected:
            with open(tmpdirname + "/test.pdf", "r", encoding="utf-8") as truth:
                assert expected.read() == truth.read()



"""Tests for coBib's EditCommand."""
# pylint: disable=unused-argument

from __future__ import annotations

import tempfile
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Type

import pytest
from typing_extensions import override

from cobib.commands import EditCommand
from cobib.config import Event, config
from cobib.database import Database
from cobib.utils.rel_path import RelPath

from .command_test import CommandTest

if TYPE_CHECKING:
    import cobib.commands


class TestEditCommand(CommandTest):
    """Tests for coBib's EditCommand.

    Note: in order to be able to test this command to at least a minimal degree, the test
    configuration sets the "editor" to be "cat". Thus, no changes will ever actually be made through
    this "editor".
    Nonetheless, this allows us to test the case where no changes are made (obviously), while also
    being able to test that the EditCommand actually writes to the database when (e.g.) the `--add`
    keyword argument is being used.
    """

    @override
    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        return EditCommand

    @staticmethod
    def _assert(
        changes: bool, logs: Optional[List[Tuple[str, int, str]]] = None, label: str = "dummy"
    ) -> None:
        """Common assertion utility method.

        Args:
            changes: whether actual changes were applied.
            logs: the list of logged messages.
            label: the label of the edited `Entry`.
        """
        if changes:
            if logs is not None:
                assert ("cobib.commands.edit", 20, f"'{label}' was successfully edited.") in logs

            with open(config.database.file, "r", encoding="utf-8") as file:
                lines = file.readlines()
                dummy_start = lines.index("dummy:\n")
                assert dummy_start > 0
                assert lines[dummy_start - 1] == "---\n"
                assert lines[dummy_start + 1] == "  ENTRYTYPE: article\n"
                assert lines[dummy_start + 2] == "...\n"
        else:
            if logs is not None:
                assert ("cobib.commands.edit", 20, "No changes detected.") in logs

    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False}],
            [{"git": True}],
        ],
        indirect=["setup"],
    )
    @pytest.mark.parametrize(
        ["args", "changes"],
        [
            [["einstein"], False],
            [["-a", "dummy"], True],
        ],
    )
    def test_command(
        self, setup: Any, caplog: pytest.LogCaptureFixture, args: List[str], changes: bool
    ) -> None:
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
            args: the arguments to pass to the command.
            changes: whether actual changes will be applied.
        """
        git = setup.get("git", False)

        EditCommand(*args).execute()

        true_log = [rec for rec in caplog.record_tuples if rec[0] == "cobib.commands.edit"]

        # check common log
        expected_log = [
            ("cobib.commands.edit", 10, "Starting Edit command."),
            ("cobib.commands.edit", 10, "Creating temporary file."),
            ("cobib.commands.edit", 10, 'Starting editor "cat".'),
            ("cobib.commands.edit", 10, "Editor finished successfully."),
        ]

        assert true_log[0:4] == expected_log
        self._assert(changes=changes, logs=true_log)

        if git and changes:
            # assert the git commit message
            self.assert_git_commit_message(
                "edit", {"label": args[-1], "add": "-a" in args, "preserve_files": None}
            )

    def test_ignore_add_if_label_exists(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test that the `add` argument is ignored if the label already exists.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        EditCommand("-a", "einstein").execute()
        assert (
            "cobib.commands.edit",
            30,
            "Entry 'einstein' already exists! Ignoring the `--add` argument.",
        ) in caplog.record_tuples

    @pytest.mark.parametrize("preserve_files", [None, True, False])
    @pytest.mark.parametrize("config_overwrite", [True, False])
    def test_rename_associated_file(
        self, setup: Any, preserve_files: bool, config_overwrite: bool
    ) -> None:
        """Test removing associated files.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            preserve_files: argument to `EditCommand`.
            config_overwrite: what to overwrite `config.commands.edit.preserve_files` with.
        """
        config.commands.edit.preserve_files = config_overwrite

        should_preserve = config_overwrite
        if preserve_files is not None:
            should_preserve = preserve_files

        try:
            config.commands.edit.editor = "sed -i 's/einstein:/dummy:/'"

            with tempfile.TemporaryDirectory() as tmpdirname:
                path = RelPath(tmpdirname + "/einstein.pdf")
                open(  # pylint: disable=consider-using-with
                    path.path, "w", encoding="utf-8"
                ).close()

                Database()["einstein"].file = str(path)

                args = ["einstein"]
                if preserve_files is not None:
                    args.insert(2, f"--{'' if preserve_files else 'no-'}preserve-files")
                EditCommand(*args).execute()
                assert "dummy" in Database().keys()

                target = RelPath(tmpdirname + "/dummy.pdf")

                if should_preserve:
                    assert path.path.exists()
                else:
                    assert target.path.exists()
        finally:
            config.defaults()

    def test_warning_missing_label(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test warning for missing label.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        EditCommand("dummy").execute()
        assert (
            "cobib.commands.edit",
            40,
            "No entry with the label 'dummy' could be found.\n"
            "Use `--add` to add a new entry with this label.",
        ) in caplog.record_tuples

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["setup"],
        [
            [{"git": False}],
        ],
        indirect=["setup"],
    )
    async def test_cmdline(self, setup: Any, monkeypatch: pytest.MonkeyPatch) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
        """
        await self.run_module(monkeypatch, "main", ["cobib", "edit", "-a", "dummy"])
        self._assert(changes=True, logs=None)

    def test_event_pre_edit_command(self, setup: Any) -> None:
        """Tests the PreEditCommand event."""

        @Event.PreEditCommand.subscribe
        def hook(command: EditCommand) -> None:
            command.largs.add = True
            command.largs.label = "dummy"

        assert Event.PreEditCommand.validate()

        EditCommand("einstein").execute()

        self._assert(changes=True)

    def test_event_post_edit_command(self, setup: Any) -> None:
        """Tests the PostEditCommand event."""

        @Event.PostEditCommand.subscribe
        def hook(command: EditCommand) -> None:
            command.new_entry.data["tags"] = "test"

        assert Event.PostEditCommand.validate()

        EditCommand("-a", "dummy").execute()

        assert Database()["dummy"].data["tags"] == "test"



"""Tests for coBib's ShowCommand."""
# pylint: disable=unused-argument

from __future__ import annotations

from itertools import zip_longest
from typing import TYPE_CHECKING, Any, List, Type

import pytest
from rich.syntax import Syntax
from typing_extensions import override

from cobib.commands import ShowCommand
from cobib.config import Event

from .. import get_resource
from .command_test import CommandTest

if TYPE_CHECKING:
    import cobib.commands


class TestShowCommand(CommandTest):
    """Tests for coBib's ShowCommand."""

    @override
    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        return ShowCommand

    def _assert(self, output: List[str]) -> None:
        """Common assertion utility method.

        Args:
            output: the actual output of the command.
        """
        with open(get_resource("example_literature.bib"), "r", encoding="utf-8") as expected:
            # we use zip_longest to ensure that we don't have more than we expect
            for line, truth in zip_longest(output, expected):
                if not line:
                    continue
                assert line == truth.strip("\n")

    def test_command(self, setup: Any) -> None:  # type: ignore
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """
        cmd = ShowCommand("einstein")
        cmd.execute()
        self._assert(cmd.entry_str.split("\n"))

    def test_render_porcelain(self, setup: Any) -> None:
        """Test the porcelain rendering.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """
        cmd = ShowCommand("einstein")
        cmd.execute()
        out = cmd.render_porcelain()
        self._assert(out)

    def test_render_rich(self, setup: Any) -> None:
        """Test the rich rendering.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """
        cmd = ShowCommand("einstein")
        cmd.execute()
        renderable = cmd.render_rich()
        assert isinstance(renderable, Syntax)
        assert renderable.code == cmd.entry_str
        # pylint: disable=protected-access
        assert renderable._lexer == "bibtex"

    def test_warning_missing_label(self, setup: Any, caplog: pytest.LogCaptureFixture) -> None:
        """Test warning for missing label.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        ShowCommand("dummy").execute()
        assert (
            "cobib.commands.show",
            40,
            "No entry with the label 'dummy' could be found.",
        ) in caplog.record_tuples

    @pytest.mark.asyncio
    async def test_cmdline(
        self, setup: Any, monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
    ) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
            capsys: the built-in pytest fixture.
        """
        await self.run_module(monkeypatch, "main", ["cobib", "show", "einstein"])
        self._assert(capsys.readouterr().out.strip().split("\n"))

    def test_event_pre_show_command(self, setup: Any) -> None:
        """Tests the PreShowCommand event."""

        @Event.PreShowCommand.subscribe
        def hook(command: ShowCommand) -> None:
            command.largs.label = "einstein"

        assert Event.PreShowCommand.validate()

        cmd = ShowCommand("knuthwebsite")
        cmd.execute()
        out = cmd.render_porcelain()
        self._assert(out)

    def test_event_post_show_command(self, setup: Any) -> None:
        """Tests the PostShowCommand event."""

        @Event.PostShowCommand.subscribe
        def hook(command: ShowCommand) -> None:
            command.entry_str = "Hello world!"

        assert Event.PostShowCommand.validate()

        cmd = ShowCommand("knuthwebsite")
        cmd.execute()
        out = cmd.render_porcelain()
        assert out == ["Hello world!"]



"""coBib's Rich-Textual progress widget.

This widget is a bit of a Frankenstein project as it wraps rich's `rich.progress.Progress` object
and makes sure that upon advancing a task, the `textual.widgets.Static` widget used to display it in
the TUI gets updated. This was favored over using textual's `textual.widgets.ProgressBar` to
re-implement rich's more flexible `rich.progress.Progress` displaying capabilities.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

from threading import RLock
from time import monotonic

from rich.progress import Progress as RichProgress
from rich.progress import ProgressColumn, Task, TaskID
from textual.widgets import Static
from typing_extensions import override


class Progress(  # type: ignore[misc]
    Static, RichProgress, can_focus=False, can_focus_children=False
):
    """coBib's Frankenstein Rich-Textual progress widget."""

    DEFAULT_CSS = """
        Progress {
            layout: horizontal;
            width: 100%;
            height: 1;
        }
    """

    def __init__(
        self,
        *columns: str | ProgressColumn,
    ) -> None:
        """Initializes the Frankenstein Rich-Textual progress widget.

        Args:
            *columns: the columns to be reported. If this is empty, it will fall back to
                `get_default_columns`.
        """
        super().__init__(id="live")

        self.columns = columns or self.get_default_columns()

        self._lock = RLock()
        self._tasks: dict[TaskID, Task] = {}
        self._task_index: TaskID = TaskID(0)

        self.speed_estimate_period = 30.0
        self.get_time = monotonic

    @override
    def advance(self, task_id: TaskID, advance: float = 1) -> None:
        super().advance(task_id, advance)
        self.renderable = self.make_tasks_table(self.tasks)
        self.refresh()

    @override
    def start(self) -> None:
        return

    @override
    def stop(self) -> None:
        return



"""coBib's UI components.

This module provides various UI-related components such as widgets and other utilities.

.. warning::

   This module makes no API stability guarantees! With the `cobib.ui.tui.TUI` being based on
   [`textual`](https://textual.textualize.io/) which is still in very early stages of its
   development, breaking API changes in this module might be released as part of coBib's feature
   releases. You have been warned.
"""

from .argument_parser import ArgumentParser as ArgumentParser
from .entry_view import EntryView as EntryView
from .help_screen import HelpScreen as HelpScreen
from .input_screen import InputScreen as InputScreen
from .list_view import ListView as ListView
from .main_content import MainContent as MainContent
from .motion_key import MotionKey as MotionKey
from .popup import Popup as Popup
from .popup_logging_handler import PopupLoggingHandler as PopupLoggingHandler
from .popup_panel import PopupPanel as PopupPanel
from .progress import Progress as Progress
from .prompt import Prompt as Prompt
from .search_view import SearchView as SearchView
from .selection_filter import SelectionFilter as SelectionFilter



"""coBib's Test Suite.

The test suite is organized into submodules which mirror the layout of the source code. As such,
each file in the source should have a corresponding file with unittests.
In most cases, the unittests will be organized into classes which derive from a common base test
class at the submodule level.

To run the test suite with coverage reporting, use:
```
tox -e coverage
```
For quick testing of individual files, it is easier to use:
```
pytest <path/to/test/file.py>
```
Check out `pytest --help` for more options.
"""

from pathlib import Path, PurePath
from typing import List, Optional

from .cmdline_test import CmdLineTest


def get_resource(filename: str, path: Optional[str] = None) -> str:
    """A utility method to get the absolute path to a resource in the test suite.

    Args:
        filename: the name of the file to get.
        path: an optional path relative to the root of the test suite.

    Returns:
        The absolute path of the file.
    """
    root = Path(__file__).parent
    full_path = root if path is None else root / Path(path)
    return str(full_path / filename)


class MockStdin:
    """A mock object to replace `sys.stdin`."""

    # pylint: disable=missing-function-docstring

    def __init__(self, string: Optional[List[str]] = None) -> None:
        # noqa: D107
        if string is None:
            string = []
        self.string = string + ["\n"]

    def readline(self) -> str:
        # noqa: D102
        return self.string.pop(0)



"""coBib's parsers.

coBib provides a variety of parsers which handle the translation from (and to) specific sources into
`cobib.database.Entry` instances.
The abstract interface which should be implemented is defined by the `cobib.parsers.base_parser`.
"""

from .arxiv import ArxivParser as ArxivParser
from .bibtex import BibtexParser as BibtexParser
from .doi import DOIParser as DOIParser
from .isbn import ISBNParser as ISBNParser
from .url import URLParser as URLParser
from .yaml import YAMLParser as YAMLParser



"""coBib command test class."""

from __future__ import annotations

import json
import logging
import os
import shlex
import subprocess
import tempfile
from abc import abstractmethod
from pathlib import Path
from shutil import copyfile, rmtree
from typing import TYPE_CHECKING, Any, Dict, Generator, Optional, Type

import pytest

from cobib.config import config
from cobib.database import Database

from .. import get_resource
from ..cmdline_test import CmdLineTest

TMPDIR = Path(tempfile.gettempdir()).resolve()

if TYPE_CHECKING:
    import _pytest.fixtures

    import cobib.commands


class CommandTest(CmdLineTest):
    """The base class for coBib's command test classes."""

    COBIB_TEST_DIR = TMPDIR / "cobib_test"
    """Path to the temporary coBib test directory."""

    COBIB_TEST_DIR_GIT = COBIB_TEST_DIR / ".git"
    """Path to the temporary coBib test git directory."""

    @abstractmethod
    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        """Get the command tested by this class.

        Returns:
            The command to be tested by this class.
        """

    @abstractmethod
    def test_command(self) -> None:
        """Test the command itself."""

    @pytest.fixture
    def setup(self, request: _pytest.fixtures.SubRequest) -> Generator[Dict[str, Any], None, None]:
        """Setup testing environment.

        This method takes care of setting up the testing configuration. This involves:
            - using `cat` for `config.commands.edit.editor` and `config.commands.open.command`
            - setting up a testing database based (if requested by `request.param["database"]`)
            - initializing git tracking (if requested by `request.param["git"]`)

        After yielding to the actual test, this setup also gets torn down.

        Args:
            request: a pytest sub-request providing access to nested parameters.

        Yields:
            The internally used parameters for potential later re-use during the actual test.
        """
        if not hasattr(request, "param"):
            # use default settings
            request.param = {"git": False, "database": True}

        # use temporary config
        config.commands.delete.confirm = False
        config.commands.edit.editor = "cat"
        config.commands.open.command = "cat"
        config.database.file = str(self.COBIB_TEST_DIR / "database.yaml")
        config.database.git = request.param.get("git", False)
        config.utils.file_downloader.default_location = "/tmp"
        config.logging.version = None

        # load database
        if request.param.get("database", True):
            self.COBIB_TEST_DIR.mkdir(parents=True, exist_ok=True)
            copyfile(get_resource("example_literature.yaml"), config.database.file)
            Database().read()
            if request.param.get("git", True):
                root = self.COBIB_TEST_DIR
                msg = "Initial commit"
                commands = [
                    f"cd {root}",
                    "git init",
                    "git add -- database.yaml",
                    f"git commit --no-gpg-sign --quiet --message {shlex.quote(msg)}",
                ]
                os.system("; ".join(commands))

        yield request.param

        # clean up file system
        try:
            os.remove(config.database.file)
            if request.param.get("git", False):
                rmtree(self.COBIB_TEST_DIR_GIT)
            # remove all handlers from root logger as to not accumulate them during the test suite
            logging.getLogger().handlers = []
        except FileNotFoundError:
            pass

        # clean up database
        if request.param.get("database", True):
            Database().clear()

        # clean up config
        config.defaults()

    def assert_git_commit_message(
        self, command: str, args: Optional[Dict[str, Any]] = None, msg: Optional[str] = None
    ) -> None:
        """Assert the last auto-generated git commit message.

        Args:
            command: the last command type which ran.
            args: the arguments which were passed to the command.
            msg: an optional commit message which overrules the auto-generated message from the
                 previous arguments.
        """
        # get last commit message
        with subprocess.Popen(
            ["git", "-C", self.COBIB_TEST_DIR, "show", "--format=format:%B", "--no-patch", "HEAD"],
            stdout=subprocess.PIPE,
        ) as proc:
            message, _ = proc.communicate()
            # decode it
            split_msg = message.decode("utf-8").split("\n")
            if split_msg is None:
                return

            if msg is not None:
                assert msg == "\n".join(split_msg)
                return

            # assert subject line
            assert f"Auto-commit: {command.title()}Command" in split_msg[0]

            if args is not None:
                # assert args
                args_str = json.dumps(args, indent=2, default=str)
                for ref, truth in zip(args_str.split("\n"), split_msg[2:]):
                    assert ref == truth

    def test_handle_argument_error(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test handling of ArgumentError.

        Args:
            caplog: the built-in pytest fixture.
        """
        command_cls = self.get_command()
        try:
            command_cls("--dummy").execute()
        except SystemExit:
            pass
        name = command_cls.name
        for source, level, message in caplog.record_tuples:
            if ("cobib.commands.base_command", logging.ERROR) == (
                source,
                level,
            ) and f"Error: {name}: error:" in message:
                break
        else:
            pytest.fail("No Error logged from ArgumentParser.")



"""coBib's entry viewer widget.

This widget gets used to display the result of `cobib.commands.show.ShowCommand.render_textual`.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

from textual.app import ComposeResult
from textual.containers import VerticalScroll
from textual.widgets import Static
from typing_extensions import override


class EntryView(VerticalScroll):
    """coBib's entry viewer widget."""

    DEFAULT_CSS = """
        EntryView {
            height: 1fr;
            width: 1fr;
        }
    """

    @override
    def compose(self) -> ComposeResult:
        yield Static()



"""coBib's UI base class.

This class handles the global command-line options which are available across all commands,
including the TUI.
"""

import argparse
import logging
from abc import abstractmethod
from typing import Any

from cobib.config import config
from cobib.ui.components import ArgumentParser
from cobib.utils.logging import get_file_handler, get_stream_handler

LOGGER = logging.getLogger(__name__)


class UI:
    """The UI base class.

    The following global arguments can be parsed:

        * `-v`, `--verbose`: increase the logging verbosity for every time this argument is given.
        * `-p`, `--porcelain`: switches the output to porcelain mode.
        * `-l`, `--logfile`: provides the path to an alternative logging file, overwriting the
            `cobib.config.config.LoggingConfig.logfile` setting.
        * `-c`, `--config`: provides the path to an alternative configuration file.
    """

    parser: ArgumentParser

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initializes a UI object.

        The function signature is left purposefully vague to allow arbitrarily complex UI subclasses
        built on top of this.

        Args:
            *args: arbitrary positional arguments.
            **kwargs: arbitrary keyword arguments.
        """
        self.root_logger = logging.getLogger()
        """Provides unified access to the root `logging.Logger`."""

        self._stream_handler = get_stream_handler()

        self.root_logger.setLevel("DEBUG")
        self.root_logger.addHandler(self._stream_handler)

        super().__init__(*args, **kwargs)

    def init_argument_parser(self, **kwargs: Any) -> None:
        """Initializes the `argparse.ArgumentParser` for global command-line arguments.

        This method needs to called by a subclass manually (preferably) during its `__init__`. The
        keyword arguments can be used to add additional information to the `ArgumentParser`, which
        can be used to inject information into the `--help` output.

        Args:
            **kwargs: arbitrary keyword arguments passed on to the `ArgumentParser` constructor.
        """
        self.parser = ArgumentParser(**kwargs)
        self.parser.add_argument("-v", "--verbose", action="count", default=0)
        self.parser.add_argument(
            "-p",
            "--porcelain",
            action="store_true",
            help="switches the output to porcelain mode (meant for parsing/testing)",
        )
        self.parser.add_argument(
            "-l", "--logfile", type=argparse.FileType("w"), help="Alternative log file"
        )
        self.parser.add_argument(
            "-c", "--config", type=argparse.FileType("r"), help="Alternative config file"
        )
        self.add_extra_parser_arguments()

    def add_extra_parser_arguments(self) -> None:
        """A hook to register additional command-line arguments.

        Subclasses can overwrite this method to add additional arguments to the `ArgumentParser`.
        This method is internally called during `init_argument_parser`.
        """

    def parse_args(self) -> argparse.Namespace:
        """Parses the provided command-line arguments.

        This method does not take any arguments because it directly gathers them from the
        command-line.

        Returns:
            The parsed arguments.
        """
        arguments = self.parser.parse_args()

        if arguments.logfile:
            LOGGER.info("Switching to FileHandler logger in %s", arguments.logfile.name)
            file_handler = get_file_handler(
                "DEBUG" if arguments.verbose > 1 else "INFO", logfile=arguments.logfile.name
            )
            self.root_logger.addHandler(file_handler)

        # set logging verbosity level
        if arguments.verbose == 1:
            self._stream_handler.setLevel(logging.INFO)
            LOGGER.info("Logging level set to INFO.")
        elif arguments.verbose > 1:
            self._stream_handler.setLevel(logging.DEBUG)
            LOGGER.info("Logging level set to DEBUG.")

        # load configuration
        config.load(arguments.config)

        return arguments

    @abstractmethod
    def run(self) -> None:
        """Runs the actual UI instance."""



"""coBib's path utility."""

from pathlib import Path
from typing import Any, Union


class RelPath:
    """The RelPath object.

    This object is a simple wrapper of a `pathlib.Path` object which ensures that a path is, if
    possible, relative to the user's home directory or an absolute path.
    This path does *not* get expanded when converting the object to a `str`, which happens during
    storage in the database.
    *Only* when accessing other attributes, will they be forwarded to a `pathlib.Path` instance of
    the fully-resolved, absolute path.
    """

    HOME = Path.home()
    """The path of the user's home directory."""

    def __init__(self, path: Union[str, Path]) -> None:
        """Initializes the path.

        This will first expand and fully resolve the given path and store it internally as a path
        relative to the user's home directory (if possible) or as an absolute path.

        Args:
            path: the path to store.
        """
        full_path = Path(path).expanduser().resolve()
        try:
            self._path = "~" / full_path.relative_to(self.HOME)
        except ValueError:
            self._path = full_path

    def __getattr__(self, attr: str) -> Any:
        """Gets an attribute from the internal path.

        Args:
            attr: the attribute to get.

        Returns:
            The attribute's value.
        """
        return getattr(self.path, attr)

    def __str__(self) -> str:
        """Transforms the internal path to a string.

        This is the only method which will not automatically expand the internal path to its
        fully-resolved, absolute version.
        """
        return str(self._path)

    @property
    def path(self) -> Path:
        """The fully-resolved, absolute path."""
        return self._path.expanduser().resolve()



"""coBib's importers.

coBib provides various "importers" which handle the migration from other bibliography managers.
The abstract interface which should be implemented is defined in `cobib.importers.base_importer`.
"""

from .zotero import ZoteroImporter as ZoteroImporter



"""coBib's ArgumentParser.

This is merely a simple wrapper around the builtin `argparse.ArgumentParser` in order to bring the
[`exit_on_error`](https://docs.python.org/3/library/argparse.html#exit-on-error) behavior to Python
versions lower than 3.9.

.. warning::

   Since this is merely a utility to support Python 3.8, once support for that will be dropped, this
   class will be removed without further notice.
"""

import argparse
from typing import NoReturn, Optional


class ArgumentParser(argparse.ArgumentParser):
    """Wrapper of the `argparse.ArgumentParser` to allow catching of error messages.

    Note, this class will be removed once Python 3.9 becomes the minimal supported version as it
    added the [`exit_on_error`](https://docs.python.org/3/library/argparse.html#exit-on-error)
    keyword argument.
    """

    # TODO: once Python 3.9 becomes the default, make use of the exit_on_error argument.

    def exit(self, status: int = 0, message: Optional[str] = None) -> NoReturn:
        """Overwrite the exit method to raise an error rather than exit.

        Args:
            status: the status code. If non-zero, an `argparse.ArgumentError` will be raised.
            message: the message of the error.

        Raises:
            An `argparse.ArgumentError`.
        """
        if status:
            raise argparse.ArgumentError(None, f"Error: {message}")
        super().exit(status, message)  # pragma: no cover



"""coBib GitCommit event tests."""
# pylint: disable=unused-argument

from __future__ import annotations

import argparse
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional, Type

import pytest
from rich.console import Console
from rich.prompt import PromptBase, PromptType
from textual.app import App
from typing_extensions import override

from cobib.commands.base_command import Command
from cobib.config import Event, config
from cobib.utils.rel_path import RelPath

from .command_test import CommandTest

if TYPE_CHECKING:
    import cobib.commands


class DummyCommand(Command):
    """A dummy command to test the GitCommit events."""

    name = "dummy"

    @override
    def __init__(
        self,
        *args: str,
        console: Console | App[None] | None = None,
        prompt: Type[PromptBase[PromptType]] | None = None,
    ) -> None:
        # pylint: disable=super-init-not-called
        self.largs = argparse.Namespace()

    @override
    @classmethod
    def init_argparser(cls) -> None:
        pass

    def execute(self) -> None:
        """Does nothing but generate a dummy git commit.

        Args:
            args: a sequence of additional arguments used for the execution.
            out: the output IO stream. This defaults to `sys.stdout`.
        """
        with open(config.database.file, "a", encoding="utf-8") as file:
            file.write("dummy")

        self.git()


class TestGitCommitEvent(CommandTest):
    """Tests for the automatic git-commit related events."""

    @override
    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        return DummyCommand

    @override
    def test_command(self) -> None:
        pytest.skip("The dummy command has no actual command.")

    @override
    def test_handle_argument_error(self, caplog: pytest.LogCaptureFixture) -> None:
        pytest.skip("The dummy command has no argument parser.")

    @pytest.mark.parametrize("setup", [{"git": True}], indirect=["setup"])
    def test_event_pre_git_commit(self, setup: Any) -> None:
        """Test the PreGitCommit event.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """

        @Event.PreGitCommit.subscribe
        def hook(msg: str, args: Optional[Dict[str, Any]] = None) -> Optional[str]:
            return "Hello world!"

        assert Event.PreGitCommit.validate()

        DummyCommand().execute()

        self.assert_git_commit_message("dummy", msg="Hello world!\n")

    @pytest.mark.parametrize("setup", [{"git": True}], indirect=["setup"])
    def test_event_post_git_commit(self, setup: Any) -> None:
        """Test the PostGitCommit event.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """

        @Event.PostGitCommit.subscribe
        def hook(root: Path, file: Path) -> None:
            file.unlink()

        assert Event.PostGitCommit.validate()

        DummyCommand().execute()

        assert not RelPath(config.database.file).path.exists()



"""Tests for coBib's shell helper functions."""

import logging
import os
import subprocess
import tempfile
from itertools import zip_longest
from pathlib import Path
from shutil import copyfile, rmtree
from typing import Generator, List, Optional, Set, Union

import pytest

from cobib.config import config
from cobib.database import Database
from cobib.utils import shell_helper
from cobib.utils.rel_path import RelPath

from .. import get_resource
from ..cmdline_test import CmdLineTest

LOGGER = logging.getLogger()


@pytest.fixture(autouse=True)
def ensure_logging_not_altered() -> Generator[None, None, None]:
    """Ensures that the logging framework remains unaltered.

    Reference:
        https://github.com/pytest-dev/pytest/issues/5743
    """
    before_handlers = list(LOGGER.handlers)
    yield
    LOGGER.handlers = before_handlers


class ShellHelperTest(CmdLineTest):
    """A base class for some common shell helper unit tests."""

    COMMAND: Optional[str] = None
    """The name of the shell helper command-"""

    EXPECTED: Optional[Union[str, List[str], Set[str]]] = None
    """The expected outcome."""

    def _assert(self, out: str) -> None:
        """The utility assertion method.

        Args:
            out: The captured output of the shell helper function.
        """
        if isinstance(self.EXPECTED, list):
            assert out.split() == self.EXPECTED
        elif isinstance(self.EXPECTED, set):
            assert set(out.split()) == self.EXPECTED

    def test_method(self) -> None:
        """Test the shell_helper method itself."""
        cmds = getattr(shell_helper, str(self.COMMAND))()
        self._assert("\n".join(cmds))

    @pytest.mark.asyncio
    async def test_cmdline(
        self, monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
    ) -> None:
        """Test the command-line access of the helper method.

        Args:
            monkeypatch: the built-in pytest fixture.
            capsys: the built-in pytest fixture.
        """
        await CmdLineTest.run_module(monkeypatch, "main", ["cobib", f"_{self.COMMAND}"])
        self._assert(capsys.readouterr().out)

    @pytest.mark.asyncio
    async def test_cmdline_via_main(
        self, monkeypatch: pytest.MonkeyPatch, capsys: pytest.CaptureFixture[str]
    ) -> None:
        """Test the command-line access of the helper method via the main method.

        Args:
            monkeypatch: the built-in pytest fixture.
            capsys: the built-in pytest fixture.
        """
        await CmdLineTest.run_module(monkeypatch, "main", ["cobib", f"_{self.COMMAND}"])
        self._assert(capsys.readouterr().out)


class TestListCommands(ShellHelperTest):
    """Tests for the shell helper which lists the available commands."""

    @staticmethod
    @pytest.fixture(autouse=True)
    def setup() -> None:
        """Setup debugging config.

        This fixture is automatically enabled for all tests in this class.
        """
        config.defaults()
        config.load(get_resource("debug.py"))

    COMMAND = "list_commands"
    EXPECTED = [
        "add",
        "delete",
        "edit",
        "export",
        "import",
        "init",
        "list",
        "modify",
        "open",
        "redo",
        "search",
        "show",
        "undo",
    ]


class TestListLabels(ShellHelperTest):
    """Tests for the shell helper which lists the existing labels."""

    COMMAND = "list_labels"
    EXPECTED = ["einstein", "latexcompanion", "knuthwebsite"]


class TestListFilters(ShellHelperTest):
    """Tests for the shell helper which lists the existing filters."""

    COMMAND = "list_filters"
    EXPECTED = {
        "publisher",
        "ENTRYTYPE",
        "address",
        "label",
        "journal",
        "doi",
        "year",
        "title",
        "author",
        "pages",
        "number",
        "volume",
        "url",
    }


class TestPrintExampleConfig(ShellHelperTest):
    """Tests for the shell helper which prints the example configuration."""

    COMMAND = "example_config"
    EXPECTED = get_resource("example.py", "../src/cobib/config")

    def _assert(self, out: str) -> None:
        with open(self.EXPECTED, "r", encoding="utf-8") as expected:
            for line, truth in zip_longest(out.split("\n"), expected):
                try:
                    assert line == truth.strip()
                except AttributeError:
                    # an empty string can equal no string (i.e. None)
                    assert bool(line) == bool(truth)


class TestLintDatabase(ShellHelperTest):
    """Tests for the shell helper which lints the users database."""

    COMMAND = "lint_database"
    REL_PATH = RelPath(get_resource("linting_database.yaml", "utils"))
    EXPECTED = [
        f"{REL_PATH}:5 Converted the field 'file' of entry 'dummy' to a list. You can consider "
        "storing it as such directly.",
        f"{REL_PATH}:6 Converting field 'month' of entry 'dummy' from '8' to 'aug'.",
        f"{REL_PATH}:7 Converting field 'number' of entry 'dummy' to integer: 1.",
        f"{REL_PATH}:8 Converted the field 'tags' of entry 'dummy' to a list. You can consider "
        "storing it as such directly.",
        f"{REL_PATH}:9 Converted the field 'url' of entry 'dummy' to a list. You can consider "
        "storing it as such directly.",
        f"{REL_PATH}:4 The field 'ID' of entry 'dummy' is no longer required. It will be inferred "
        "from the entry label.",
    ]

    @staticmethod
    @pytest.fixture(autouse=True)
    def setup() -> None:
        """Set linting database path.

        This fixture is automatically enabled for all tests in this class.
        """
        config.defaults()
        config.database.file = str(TestLintDatabase.REL_PATH)

    def _assert(self, out: str) -> None:
        for msg, truth in zip_longest(out.split("\n"), self.EXPECTED):
            if msg.strip() and truth:
                assert msg == truth

    def test_no_lint_warnings(self) -> None:
        """Test the case of no raised lint warnings."""
        config.load(get_resource("debug.py"))
        lint_messages = shell_helper.lint_database()
        for msg, exp in zip_longest(
            lint_messages, ["Congratulations! Your database triggers no lint messages."]
        ):
            if msg.strip() and exp:
                assert msg == exp

    @pytest.mark.parametrize("git", [False, True])
    def test_lint_auto_format(self, git: bool) -> None:
        """Test automatic lint formatter.

        Args:
            git: whether or not git-tracking should be enabled.
        """
        tmp_dir = Path(tempfile.gettempdir()).resolve()
        cobib_test_dir = tmp_dir / "cobib_lint_test"
        cobib_test_dir.mkdir(parents=True, exist_ok=True)
        cobib_test_dir_git = cobib_test_dir / ".git"

        database_file = RelPath(cobib_test_dir / "database.yaml")

        copyfile(TestLintDatabase.REL_PATH.path, database_file.path)
        config.database.file = str(database_file)
        config.database.git = git

        if git:
            commands = [
                f"cd {cobib_test_dir}",
                "git init",
                "git add -- database.yaml",
                "git commit --no-gpg-sign --quiet --message 'Initial commit'",
            ]
            os.system("; ".join(commands))

        try:
            # apply linting with formatting and check for the expected lint messages
            pre_lint_messages = shell_helper.lint_database("--format")
            expected_messages = [
                "The following lint messages have successfully been resolved:"
            ] + self.EXPECTED
            for msg, truth in zip_longest(pre_lint_messages, expected_messages):
                if msg.strip() and truth:
                    assert msg == truth.replace(str(TestLintDatabase.REL_PATH), str(database_file))

            # assert auto-formatted database
            with open(database_file.path, "r", encoding="utf-8") as file:
                with open(
                    get_resource("fixed_database.yaml", "utils"), "r", encoding="utf-8"
                ) as expected:
                    for line, truth in zip_longest(file.readlines(), expected.readlines()):
                        assert line == truth

            # assert git message
            if git:
                with subprocess.Popen(
                    [
                        "git",
                        "-C",
                        cobib_test_dir_git,
                        "show",
                        "--format=format:%B",
                        "--no-patch",
                        "HEAD",
                    ],
                    stdout=subprocess.PIPE,
                ) as proc:
                    message, _ = proc.communicate()
                    # decode it
                    split_msg = message.decode("utf-8").split("\n")
                    if split_msg is None:
                        return
                    # assert subject line
                    assert "Auto-commit: LintCommand" in split_msg[0]

            # recheck linting and assert no lint messages
            post_lint_messages = shell_helper.lint_database()
            for msg, exp in zip_longest(
                post_lint_messages, ["Congratulations! Your database triggers no lint messages."]
            ):
                if msg.strip() and exp:
                    assert msg == exp

        finally:
            rmtree(cobib_test_dir)
            Database().clear()


class TestUnifyLabels(ShellHelperTest):
    """Tests for the shell helper which unifies all database labels."""

    COMMAND = "unify_labels"
    REL_PATH = RelPath(get_resource("unifying_database.yaml", "utils"))
    EXPECTED: List[str] = [
        "[INFO] Associated files will not be preserved.",
        "[INFO] einstein: changing field 'label' from einstein to Einstein1905_a",
        "[INFO] latexcompanion: changing field 'label' from latexcompanion to Goossens1993",
        "[INFO] knuthwebsite: changing field 'label' from knuthwebsite to Knuth",
        "[INFO] Einstein_1905: changing field 'label' from Einstein_1905 to Einstein1905_b",
        "[INFO] New and previous values match. Skipping modification of entry 'Einstein1905'.",
        "[INFO] einstein_2: changing field 'label' from einstein_2 to Einstein1905_c",
        "[INFO] New and previous values match. Skipping modification of entry 'Author2021'.",
    ]

    @staticmethod
    @pytest.fixture(autouse=True)
    def setup() -> None:
        """Set linting database path.

        This fixture is automatically enabled for all tests in this class.
        """
        config.defaults()
        config.database.format.label_default = "{author.split()[1]}{year}"
        config.database.file = str(TestUnifyLabels.REL_PATH)

    def _assert(self, out: str) -> None:
        filtered = [line for line in out.split("\n") if line.startswith("[INFO]")]
        for msg, truth in zip_longest(filtered, self.EXPECTED):
            if msg.strip() and truth:
                assert msg == truth

    @pytest.mark.parametrize("git", [False, True])
    def test_unify_labels(self, git: bool) -> None:
        """Test actual changes of label unification.

        Args:
            git: whether or not git-tracking should be enabled.
        """
        tmp_dir = Path(tempfile.gettempdir()).resolve()
        cobib_test_dir = tmp_dir / "cobib_unify_label_test"
        cobib_test_dir.mkdir(parents=True, exist_ok=True)
        cobib_test_dir_git = cobib_test_dir / ".git"

        database_file = RelPath(cobib_test_dir / "database.yaml")

        copyfile(TestUnifyLabels.REL_PATH.path, database_file.path)
        config.database.file = str(database_file)
        config.database.git = git

        if git:
            commands = [
                f"cd {cobib_test_dir}",
                "git init",
                "git add -- database.yaml",
                "git commit --no-gpg-sign --quiet --message 'Initial commit'",
            ]
            os.system("; ".join(commands))

        try:
            # apply label unification
            shell_helper.unify_labels("--apply")

            # assert unified database
            with open(database_file.path, "r", encoding="utf-8") as file:
                with open(
                    RelPath(get_resource("unified_database.yaml", "utils")).path,
                    "r",
                    encoding="utf-8",
                ) as expected:
                    for line, truth in zip_longest(file.readlines(), expected.readlines()):
                        assert line == truth

            # assert git message
            if git:
                with subprocess.Popen(
                    [
                        "git",
                        "-C",
                        cobib_test_dir_git,
                        "show",
                        "--format=format:%B",
                        "--no-patch",
                        "HEAD",
                    ],
                    stdout=subprocess.PIPE,
                ) as proc:
                    message, _ = proc.communicate()
                    # decode it
                    split_msg = message.decode("utf-8").split("\n")
                    if split_msg is None:
                        return
                    # assert subject line
                    assert "Auto-commit: ModifyCommand" in split_msg[0]

        finally:
            rmtree(cobib_test_dir)
            Database().clear()



"""coBib's Import command.

This command allows you to import new entries from another bibliography manager into coBib.
This can be seen as a migration utility and, thus, you are likely to only execute this command once.
To ease the interface (and implementation), this process of adding new entries to your database is
separated from the `cobib.commands.add.AddCommand`.

To support various bibliography managers as sources for this command, their code is split out into a
separate module, `cobib.importers`.
The various backends are registered (at runtime) in a *mutually exclusive* group of keyword
arguments, which you can use like so:
```
cobib import --zotero
```

.. note::
   Since this command adds new entries to the database, its outcome can be affected by your
   `cobib.config.config.DatabaseConfig` settings. In particular, pay attention to the
   `cobib.config.config.EntryStringifyConfig` settings which affect how entries are converted
   to/from strings. In particular, the following setting will affect how multiple files are split
   into a list of files:
   ```
   config.database.stringify.list_separator.file = ", "
   ```
   The above will separate file paths using `, ` but if you use a different separator (for example
   `;`) be sure to update this setting accordingly.

### Additional Options

Since v4.1.0 you can suppress the automatic download of attachments via the
`cobib.config.config.ImportCommandConfig.skip_download` setting. It defaults to `False` meaning that
the attachments will be downloaded
If you want to manually overwrite the configuration setting you can do so with the `--skip-download`
and `--force-download` arguments, respectively.
I.e. the following will **not** download attachments:
```
cobib import --skip-download --zotero
```
While the next command will always download attachments:
```
cobib import --force-download --zotero
```

The various importers may take even more command line arguments. Please check out their
documentation at `cobib.importers` for more details.

### TUI

You can also trigger this command from the `cobib.ui.tui.TUI`.
By default, it is bound to the `i` key which will drop you into the prompt where you can type out a
normal command-line command:
```
:import <arguments go here>
```
"""

from __future__ import annotations

import argparse
import inspect
import logging
from collections import OrderedDict
from typing import Dict, List, Type

from rich.console import Console
from rich.prompt import PromptBase, PromptType
from textual.app import App
from typing_extensions import override

from cobib import importers
from cobib.config import Event, config
from cobib.database import Database, Entry

from .base_command import ArgumentParser, Command

LOGGER = logging.getLogger(__name__)


class ImportCommand(Command):
    """The ImportCommand.

    This command can parse the following arguments:

        * `--skip-download`: skips the automatic download of attached files (like PDFs).
        * `--force-download`: forces the automatic download of attached files (like PDFs).
        * in addition to the options above, a *mutually exclusive group* of keyword arguments for
          all available `cobib.importers` are registered at runtime. Please check the output of
          `cobib import --help` for the exact list.
        * finally, you can add another set of positional arguments (preceded by `--`) which will be
          passed on to the chosen importer. For more details see for example
          `cobib import --zotero -- --help`.
    """

    name = "import"

    _avail_importers = {
        cls.name: cls for _, cls in inspect.getmembers(importers) if inspect.isclass(cls)
    }
    """The available importers."""

    @override
    def __init__(
        self,
        *args: str,
        console: Console | App[None] | None = None,
        prompt: Type[PromptBase[PromptType]] | None = None,
    ) -> None:
        super().__init__(*args, console=console, prompt=prompt)

        self.new_entries: Dict[str, Entry] = OrderedDict()
        """An `OrderedDict` mapping labels to `cobib.database.Entry` instances which were imported
        by this command."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = ArgumentParser(prog="import", description="Import subcommand parser.")
        skip_download_group = parser.add_mutually_exclusive_group()
        skip_download_group.add_argument(
            "--skip-download",
            action="store_true",
            default=None,
            help="skip the automatic download of encountered PDF attachments",
        )
        skip_download_group.add_argument(
            "--force-download",
            dest="skip_download",
            action="store_false",
            default=None,
            help="force the automatic download of encountered PDF attachments",
        )
        parser.add_argument(
            "importer_arguments",
            nargs="*",
            help="You can pass additional arguments to the chosen importer. To ensure this works as"
            " expected you should add the pseudo-argument '--' before the remaining arguments.",
        )
        group_import = parser.add_mutually_exclusive_group()
        for name in cls._avail_importers.keys():
            try:
                group_import.add_argument(f"--{name}", action="store_true", help=f"{name} importer")
            except argparse.ArgumentError:
                continue
        cls.argparser = parser

    @override
    async def execute(self) -> None:  # type: ignore[override]
        # pylint: disable=invalid-overridden-method
        LOGGER.debug("Starting Import command.")

        Event.PreImportCommand.fire(self)

        imported_entries: List[Entry] = []

        skip_download = config.commands.import_.skip_download
        if self.largs.skip_download is not None:
            skip_download = self.largs.skip_download
        LOGGER.info(
            "Associated files will%s be downloaded from the imported library.",
            "" if skip_download else " not",
        )

        for name, cls in ImportCommand._avail_importers.items():
            enabled = getattr(self.largs, name, False)
            if not enabled:
                continue
            LOGGER.debug("Importing entries from %s.", name)
            imported_entries = await cls(
                *self.largs.importer_arguments, skip_download=skip_download
            ).fetch()
            break

        bib = Database()
        existing_labels = set(bib.keys())

        for entry in imported_entries:
            # check if label already exists
            if entry.label in existing_labels:
                msg = (
                    f"The label '{entry.label}' already exists. It will be disambiguated based on "
                    "the configuration option: config.database.format.label_suffix"
                )
                LOGGER.warning(msg)
                new_label = bib.disambiguate_label(entry.label, entry)
                entry.label = new_label

            bib.update({entry.label: entry})
            existing_labels.add(entry.label)
            self.new_entries[entry.label] = entry

        Event.PostImportCommand.fire(self)
        bib.update(self.new_entries)

        bib.save()



"""coBib's Zotero importer.

This importer handles migrating from [Zotero][1] to coBib.
Normally, you would only need to trigger this migration once, but for convenience, coBib will store
the OAuth authentication tokens provided by the Zotero API in its cache (whose location is
configurable via `cobib.config.config.LoggingConfig.cache`).

The importer is registered under the `--zotero` command-line argument of the
`cobib.commands.import_.ImportCommand`. Thus, you can trigger it like so:
```
cobib import --zotero
```

### Additional Options

You can provide some additional arguments for the Zotero importer.

First of all, you can disable the cache. This is useful if you want to avoid caching any
authentication tokens or if you want to re-trigger the OAuth authentication procedure.
```
cobib import --zotero -- --no-cache
```

You can also provide a custom Zotero user ID via the command line. If you do so, OAuth
authentication will not be triggered. The import will work normally, if the library associated with
the provided user ID is publicly accessible. Otherwise, you will also need to provide your own
Zotero API key (see next paragraph).
```
cobib import --zotero -- --user-id <user ID>
```

As mentioned above, you can provide your own Zotero API key if you do not want coBib to initiate an
OAuth authentication process for you. This argument only takes affect if you also provide your own
Zotero user ID.
```
cobib import --zotero -- --user-id <user ID> --api-key <API key>
```

[1]: https://www.zotero.org/
"""

from __future__ import annotations

import json
import logging
import os
import subprocess
import sys
from typing import Dict, List

import requests
from requests_oauthlib import OAuth1Session
from typing_extensions import override

from cobib.config import Event, config
from cobib.database import Entry
from cobib.parsers import BibtexParser
from cobib.utils.file_downloader import FileDownloader
from cobib.utils.rel_path import RelPath

from .base_importer import ArgumentParser, Importer

LOGGER = logging.getLogger(__name__)


class ZoteroImporter(Importer):
    """The Zotero Importer.

    This importer can parse the following arguments:

        * `--no-cache`: disables the use of any cached OAuth tokens.
        * `--user-id`: the Zotero user ID used for API calls. You can find your user ID at
          <https://www.zotero.org/settings/keys>. Unless you also specify `--no-cache`, this value
          will be stored in coBib's internal cache. If you do not also provide an `--api-key`, you
          can only read publicly available Zotero databases.
        * `--api-key`: overwrites the user-specific Zotero API key for the coBib application. Unless
          you also specify `--no-cache`, this value will be stored in coBib's internal cache. This
          argument only takes effect if `--user-id` is also provided.
    """

    name = "zotero"

    _CLIENT_KEY = "94a2f739acd6be46df50"
    """The key given to the coBib application by the Zotero API."""
    _CLIENT_SECRET = "8d2659a20871036a462f"
    """The secret given to the coBib application by the Zotero API."""

    OAUTH_REQUEST_URL = "https://www.zotero.org/oauth/request"
    """The URL from which to initiate OAuth authentication token requests."""
    OAUTH_ACCESS_URL = "https://www.zotero.org/oauth/access"
    """The URL from which to fetch generated OAuth authentication token."""
    OAUTH_AUTHORIZE_URL = "https://www.zotero.org/oauth/authorize"
    """The URL where the user must be redirected in order to authenticate the coBib appliation."""

    @override
    def __init__(self, *args: str, skip_download: bool = False) -> None:
        super().__init__(*args, skip_download=skip_download)

        self.authentication: dict[str, str] = {}
        """The authentication dictionary used as a header during the `GET` request of the Zotero
        API."""

        self.protected_url: str = ""
        """The protected URL via which the Zotero API gets accessed."""

        self.imported_entries: list[Entry] = []
        """A list of `cobib.database.Entry` objects which were imported by this importer."""

    @override
    @classmethod
    def init_argparser(cls) -> None:
        parser = ArgumentParser(prog="zotero", description="Zotero migration parser.")
        parser.add_argument(
            "--no-cache", action="store_true", help="disable use of cached OAuth tokens"
        )
        parser.add_argument(
            "--api-key", type=str, help="the user-specific Zotero API key for the coBib application"
        )
        parser.add_argument("--user-id", type=str, help="the Zotero user ID used for API calls")

        cls.argparser = parser

    @override
    async def fetch(self) -> List[Entry]:  # type: ignore[override]
        # pylint: disable=invalid-overridden-method
        LOGGER.debug("Starting Zotero fetching.")

        if self.largs.user_id is not None:
            self.authentication["UserID"] = self.largs.user_id
            if self.largs.api_key is not None:
                self.authentication["Zotero-API-Key"] = self.largs.api_key

            if not self.largs.no_cache:
                self._store_oauth_tokens(self.authentication)
        else:
            self.authentication = self._get_authentication_tokens(self.largs.no_cache)

        user_id = self.authentication.pop("UserID")

        LOGGER.info("Requesting items from Zotero user via API v3.")
        self.protected_url = f"https://api.zotero.org/users/{user_id}/items?include=biblatex&v=3"

        Event.PreZoteroImport.fire(self)

        raw_result = requests.get(self.protected_url, headers=self.authentication, timeout=30)

        bibtex_parser = BibtexParser()

        results = json.loads(raw_result.content)

        encountered_attachments: Dict[str, Dict[str, str]] = {}

        for res in results:
            biblatex = res.get("biblatex", "").strip()
            if not bool(biblatex):
                # check if this is an attachment
                try:
                    attachment_enclosure = res.get("links", {}).get("enclosure", {})
                    LOGGER.info("Storing encountered attachment: %s", res["key"])
                    encountered_attachments[res["key"]] = {
                        "href": attachment_enclosure["href"],
                        "title": attachment_enclosure["title"],
                    }
                except KeyError:
                    pass
                continue

            LOGGER.info("Parsing encountered BibLaTeX entry: %s", res["key"])
            # biblatex contains exactly one entry so we can pop it from the OrderedDict
            _, new_entry = bibtex_parser.parse(biblatex).popitem()

            # Zotero-specific `journal` keyword handling
            new_entry.data.pop("shortjournal")
            new_entry.data["journal"] = new_entry.data.pop("journaltitle")

            # check attachment
            try:
                attachment = res.get("links", {}).get("attachment", {})
            except KeyError:
                pass

            if attachment.get("attachmentType", None) == "application/pdf":
                LOGGER.info("Queuing associated attachment for download.")
                new_entry.data["_download"] = attachment["href"].split("/")[-1]

            self.imported_entries.append(new_entry)

        for entry in self.imported_entries:  # pragma: no cover
            if "_download" not in entry.data.keys():
                continue

            key = entry.data.pop("_download")
            if self.skip_download:
                LOGGER.info("Skipping attachment download.")
                continue
            if key not in encountered_attachments:
                LOGGER.warning("Skipping unknown attachment: %s", key)
                continue

            url = encountered_attachments[key]["href"]
            filename = encountered_attachments[key]["title"]

            path = await FileDownloader().download(url, filename, headers=self.authentication)
            if path is not None:
                entry.file = str(path)  # type: ignore[assignment]

        Event.PostZoteroImport.fire(self)

        return self.imported_entries

    @staticmethod
    def _get_fresh_oauth_tokens() -> Dict[str, str]:  # pragma: no cover
        """Obtain new OAuth authentication tokens from the Zotero API.

        Returns:
            A dictionary containing the authentication information. More specifically, two values
            will be stored:
                - `Zotero-API-Key`: the API key generated by the Zotero user specifically for coBib
                - `UserID`: the Zotero user ID used during user-specific API access.
        """
        LOGGER.info("Obtaining new OAuth tokens from Zotero.")
        oauth = OAuth1Session(ZoteroImporter._CLIENT_KEY, ZoteroImporter._CLIENT_SECRET)
        LOGGER.debug("Fetching temporary request token")
        oauth.fetch_request_token(ZoteroImporter.OAUTH_REQUEST_URL)

        authorization_url = oauth.authorization_url(ZoteroImporter.OAUTH_AUTHORIZE_URL)
        LOGGER.debug("Opening user authorization URL")
        opener = "xdg-open" if sys.platform.lower() == "linux" else "open"
        with open(os.devnull, "w", encoding="utf-8") as devnull:
            subprocess.Popen(  # pylint: disable=consider-using-with
                [opener, authorization_url],
                stdout=devnull,
                stderr=devnull,
                stdin=devnull,
                close_fds=True,
            )
        redirect_response = input("Please insert the redirected URL here: ")
        LOGGER.debug("Parsing authorization redirect URL")
        oauth.parse_authorization_response(redirect_response)

        LOGGER.debug("Fetching permanent access token")
        oauth_tokens = oauth.fetch_access_token(ZoteroImporter.OAUTH_ACCESS_URL)

        authentication = {
            "Zotero-API-Key": oauth_tokens.get("oauth_token"),
            "UserID": oauth_tokens.get("userID"),
        }

        return authentication

    @staticmethod
    def _get_cached_oauth_tokens() -> Dict[str, str]:
        """Obtain the OAuth authentication tokens for the Zotero API from coBib's cache.

        Returns:
            A dictionary containing the authentication information. Refer to
            `_get_fresh_oauth_tokens` for more specific details on the dictionary contents.
        """
        LOGGER.info("Attempting to load cached OAuth tokens for Zotero.")
        cache_path = RelPath(config.logging.cache).path
        try:
            with open(cache_path, "r", encoding="utf-8") as cache:
                cached_data = json.load(cache)
        except FileNotFoundError:
            cached_data = {}
        return cached_data.get("Zotero", {})  # type: ignore[no-any-return]

    @staticmethod
    def _store_oauth_tokens(tokens: Dict[str, str]) -> None:
        """Stores the OAuth authentication tokens for the Zotero API in coBib's cache.

        Args:
            tokens: the dictionary containing the authentication information. Refer to
                `_get_fresh_oauth_tokens` for more specific details on the dictionary contents.
        """
        LOGGER.info("Storing OAuth tokens for Zotero in cache.")
        cache_path = RelPath(config.logging.cache).path
        try:
            with open(cache_path, "r", encoding="utf-8") as cache:
                cached_data = json.load(cache)
        except FileNotFoundError:
            cached_data = {}

        if "Zotero" not in cached_data.keys():
            cached_data["Zotero"] = {}
        cached_data["Zotero"].update(tokens)

        if not cache_path.parent.exists():
            cache_path.parent.mkdir(parents=True)

        with open(cache_path, "w", encoding="utf-8") as cache:
            json.dump(cached_data, cache)

    @staticmethod
    def _get_authentication_tokens(no_cache: bool = False) -> Dict[str, str]:
        """Obtain the OAuth authentication tokens for the Zotero API.

        This method will first attempt to load the tokens from the cache. If that fails, it will
        generate new tokens via the Zotero API.

        Args:
            no_cache: whether or not to use cached OAuth tokens.

        Returns:
            A dictionary containing the authentication information. Refer to
            `_get_fresh_oauth_tokens` for more specific details on the dictionary contents.
        """
        authentication: Dict[str, str] = {}

        if not no_cache:
            authentication = ZoteroImporter._get_cached_oauth_tokens()

        if authentication:
            LOGGER.info("Successfully obtained OAuth tokens for Zotero from cache.")
        else:  # pragma: no cover
            authentication = ZoteroImporter._get_fresh_oauth_tokens()
            if not no_cache:
                ZoteroImporter._store_oauth_tokens(authentication)

        return authentication



"""coBib's list results viewer widget.

This widget gets produces by `cobib.commands.list.ListCommand.render_textual`.
It subclasses textual's built-in `DataTable` widget.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

from typing import Any

from rich.text import Text
from textual.binding import Binding
from textual.coordinate import Coordinate
from textual.reactive import reactive
from textual.widgets import DataTable
from typing_extensions import override

from .motion_key import MotionKey


class ListView(DataTable[Text]):
    """coBib's list results viewer widget."""

    id = "cobib-list-view"

    cursor_type = reactive("row")

    BINDINGS = [
        Binding("j", "motion('down', 'cursor_down')", "Down", show=False),
        Binding("k", "motion('up', 'cursor_up')", "Up", show=False),
        Binding("h", "motion('left', 'cursor_left')", "Left", show=False),
        Binding("l", "motion('right', 'cursor_right')", "Right", show=False),
        Binding("down", "motion('down', 'cursor_down')", "Down", show=False),
        Binding("up", "motion('up', 'cursor_up')", "Up", show=False),
        Binding("left", "motion('left', 'cursor_left')", "Left", show=False),
        Binding("right", "motion('right', 'cursor_right')", "Right", show=False),
        # TODO: add motions for home, end, page up, and page down
    ]
    """
    | Key(s) | Description |
    | :- | :- |
    | space | Toggle the expand/collapsed space of the current item. |
    | j, down | Moves one row down. |
    | k, up | Moves one row up. |
    | h, left | Moves to the left. |
    | l, right | Moves to the right. |
    """

    DEFAULT_CSS = """
        ListView {
            height: 1fr;
        }
    """

    @override
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        # TODO: revisit this based on the outcome of https://github.com/Textualize/textual/pull/2740
        self.fixed_columns = 1
        self.zebra_stripes = True

    def action_motion(self, key: str, action: str) -> None:
        """Action to move the cursor.

        Under the hood, this delegates to the respective built-in cursor motion methods.
        However, this method also posts a `cobib.ui.components.motion_key.MotionKey` event.

        Args:
            key: the key which was pressed.
            action: the built-in action to trigger.
        """
        func = getattr(self, f"action_{action}", None)
        if func is not None:
            func()
        self.post_message(MotionKey(key))

    def get_current_label(self) -> str:
        """Gets the label of the entry currently under the cursor.

        Returns:
            The label of the entry currently under the cursor.
        """
        label = self.get_cell_at(Coordinate(self.cursor_row, 0)).plain
        return label



"""Tests for coBib's ISBNParser."""
# pylint: disable=unused-argument

import json
import logging
from typing import Dict, Optional

import pytest
import requests

from cobib.config import Event
from cobib.database import Entry
from cobib.parsers import ISBNParser

from .parser_test import ParserTest


def assert_default_test_entry(entry: Entry) -> None:
    """Asserts that the passed entry is the default testing entry.

    Args:
        entry: the entry to assert.
    """
    assert entry.label == "Lutz2013"
    assert entry.data["author"] == "Mark Lutz and David Ascher"
    assert entry.data["pages"] == 1540
    assert entry.data["title"] == "Learning Python"
    assert entry.data["year"] == 2013


class TestISBNParser(ParserTest):
    """Tests for coBib's ISBNParser."""

    def test_from_isbn(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test parsing from ISBN.

        Args:
            caplog: the built-in pytest fixture.
        """
        entries = ISBNParser().parse("978-1-449-35573-9")

        if any(
            s == "cobib.parsers.isbn" and t == logging.ERROR for s, t, _ in caplog.record_tuples
        ):
            pytest.skip("The requests API encountered an error. Skipping test.")

        entry = list(entries.values())[0]
        assert_default_test_entry(entry)

    # regression test for https://gitlab.com/cobib/cobib/-/issues/53
    def test_from_empty_isbn(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test parsing an empty ISBN.

        Args:
            caplog: the built-in pytest fixture.
        """
        entries = ISBNParser().parse("3860704443")

        if any(
            s == "cobib.parsers.isbn" and t == logging.ERROR for s, t, _ in caplog.record_tuples
        ):
            pytest.skip("The requests API encountered an error. Skipping test.")

        assert not entries
        assert entries == {}  # pylint: disable=C1803

        assert ("cobib.parsers.isbn", logging.WARNING) in [
            (source, level) for source, level, _ in caplog.record_tuples
        ]

    def test_catching_api_error(
        self, caplog: pytest.LogCaptureFixture, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Test catching API error.

        Args:
            caplog: the built-in pytest fixture.
            monkeypatch: the built-in pytest fixture.
        """

        def raise_exception(*args, **kwargs):  # type: ignore
            """Mock function to raise an Exception."""
            raise requests.exceptions.RequestException()

        monkeypatch.setattr(requests, "get", raise_exception)
        ISBNParser().parse("978-1-449-35573-9")

        assert (
            "cobib.parsers.isbn",
            logging.ERROR,
            "An Exception occurred while trying to query the ISBN: 978-1-449-35573-9.",
        ) in caplog.record_tuples

    def test_catching_decode_error(
        self, caplog: pytest.LogCaptureFixture, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Test catching json decode error.

        Args:
            caplog: the built-in pytest fixture.
            monkeypatch: the built-in pytest fixture.
        """

        def raise_exception(*args, **kwargs):  # type: ignore
            """Mock function to raise an Exception."""
            raise json.JSONDecodeError("", "", 0)

        monkeypatch.setattr(json, "loads", raise_exception)
        ISBNParser().parse("978-1-449-35573-9")

        for record in caplog.record_tuples:
            if (
                record[0] == "cobib.parsers.isbn"
                and record[1] == logging.ERROR
                and "An Exception occurred while parsing the query results:" in record[2]
            ):
                break
        else:
            pytest.fail("No Error caught by ISBNParser")

    def test_dump(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test dumping.

        Args:
            caplog: the built-in pytest fixture.
        """
        entry = Entry("dummy", {"ENTRYTYPE": "unpublished"})
        ISBNParser().dump(entry)

        assert (
            "cobib.parsers.isbn",
            logging.ERROR,
            "Cannot dump an entry as an ISBN.",
        ) in caplog.record_tuples

    def test_event_pre_isbn_parse(self, caplog: pytest.LogCaptureFixture) -> None:
        """Tests the PreISBNParse event."""

        @Event.PreISBNParse.subscribe
        def hook(string: str) -> Optional[str]:
            return "978-1-449-35573-9"

        assert Event.PreISBNParse.validate()

        entries = ISBNParser().parse("Hello world!")
        if any(
            s == "cobib.parsers.isbn" and t == logging.ERROR for s, t, _ in caplog.record_tuples
        ):
            pytest.skip("The requests API encountered an error. Skipping test.")

        entry = list(entries.values())[0]
        assert_default_test_entry(entry)

    def test_event_post_isbn_parse(self, caplog: pytest.LogCaptureFixture) -> None:
        """Tests the PostISBNParse event."""

        @Event.PostISBNParse.subscribe
        def hook(bib: Dict[str, Entry]) -> None:
            bib["Lutz2013"].data["test"] = "dummy"

        assert Event.PostISBNParse.validate()

        entries = ISBNParser().parse("978-1-449-35573-9")
        if any(
            s == "cobib.parsers.isbn" and t == logging.ERROR for s, t, _ in caplog.record_tuples
        ):
            pytest.skip("The requests API encountered an error. Skipping test.")

        entry = list(entries.values())[0]
        assert_default_test_entry(entry)
        assert entry.data["test"] == "dummy"



"""Tests for coBib's ArxivParser."""
# pylint: disable=unused-argument

import logging
from typing import Dict, Optional

import pytest
import requests

from cobib.config import Event
from cobib.database import Entry
from cobib.parsers import ArxivParser

from .parser_test import ParserTest


def assert_default_test_entry(entry: Entry) -> None:
    """Asserts that the passed entry is the default testing entry.

    Args:
        entry: the entry to assert.
    """
    entry.escape_special_chars()
    assert entry.label == "Cao2018"
    assert entry.data["archivePrefix"] == "arXiv"
    assert entry.data["arxivid"].startswith("1812.09976")
    assert (
        entry.data["author"]
        == "Yudong Cao and Jonathan Romero and Jonathan P. Olson and Matthias Degroote and "
        + "Peter D. Johnson and M{\\'a}ria Kieferov{\\'a} and Ian D. Kivlichan and Tim Menke "
        + "and Borja Peropadre and Nicolas P. D. Sawaya and Sukin Sim and Libor Veis and "
        + "Al{\\'a}n Aspuru-Guzik"
    )
    assert entry.data["doi"].startswith("10.1021/acs.chemrev.8b00803")
    assert entry.data["title"] == "Quantum Chemistry in the Age of Quantum Computing"
    assert entry.data["year"] == 2018
    assert entry.data["_download"] == "http://arxiv.org/pdf/1812.09976v2"


class TestArxivParser(ParserTest):
    """Tests for coBib's ArxivParser."""

    @pytest.mark.parametrize("query", ["1812.09976", "https://arxiv.org/abs/1812.09976"])
    def test_from_arxiv(self, query: str, caplog: pytest.LogCaptureFixture) -> None:
        """Test parsing from arXiv.

        Args:
            query: the arXiv ID or URL which to query.
            caplog: the built-in pytest fixture.
        """
        entries = ArxivParser().parse(query)

        if (
            "cobib.parsers.arxiv",
            logging.ERROR,
            "An Exception occurred while trying to query the arXiv ID: 1812.09976.",
        ) in caplog.record_tuples:
            pytest.skip("The requests API encountered an error. Skipping test.")

        entry = list(entries.values())[0]
        assert_default_test_entry(entry)

    # regression test for https://gitlab.com/cobib/cobib/-/issues/57
    def test_invalid_arxiv_id(self) -> None:
        """Test parsing an invalid arXiv ID."""
        entries = ArxivParser().parse("10.1021/acs.chemrev.8b00803")
        assert not entries
        assert entries == {}  # pylint: disable=C1803

    def test_arxiv_without_doi(self) -> None:
        """Test parsing an arXiv ID without an associated DOI."""
        entries = ArxivParser().parse("1701.08213")
        entry = list(entries.values())[0]
        assert entry.label == "Bravyi2017"
        assert entry.data["archivePrefix"] == "arXiv"
        assert entry.data["arxivid"].startswith("1701.08213")
        assert (
            entry.data["author"]
            == "Sergey Bravyi and Jay M. Gambetta and Antonio Mezzacapo and Kristan Temme"
        )
        assert entry.data["title"] == "Tapering off qubits to simulate fermionic Hamiltonians"
        assert entry.data["year"] == 2017

    def test_catching_api_error(
        self, caplog: pytest.LogCaptureFixture, monkeypatch: pytest.MonkeyPatch
    ) -> None:
        """Test catching API error.

        Args:
            caplog: the built-in pytest fixture.
            monkeypatch: the built-in pytest fixture.
        """

        def raise_exception(*args, **kwargs):  # type: ignore
            """Mock function to raise an Exception."""
            raise requests.exceptions.RequestException()

        monkeypatch.setattr(requests, "get", raise_exception)
        ArxivParser().parse("1812.0997")

        assert (
            "cobib.parsers.arxiv",
            logging.ERROR,
            "An Exception occurred while trying to query the arXiv ID: 1812.0997.",
        ) in caplog.record_tuples

    def test_dump(self, caplog: pytest.LogCaptureFixture) -> None:
        """Test dumping.

        Args:
            caplog: the built-in pytest fixture.
        """
        entry = Entry("dummy", {"ENTRYTYPE": "unpublished"})
        ArxivParser().dump(entry)

        assert (
            "cobib.parsers.arxiv",
            logging.ERROR,
            "Cannot dump an entry as an arXiv ID.",
        ) in caplog.record_tuples

    def test_event_pre_arxiv_parse(self, caplog: pytest.LogCaptureFixture) -> None:
        """Tests the PreArxivParse event."""

        @Event.PreArxivParse.subscribe
        def hook(string: str) -> Optional[str]:
            return "1812.09976"

        assert Event.PreArxivParse.validate()

        entries = ArxivParser().parse("Hello world!")
        if (
            "cobib.parsers.arxiv",
            logging.ERROR,
            "An Exception occurred while trying to query the arXiv ID: 1812.09976.",
        ) in caplog.record_tuples:
            pytest.skip("The requests API encountered an error. Skipping test.")

        entry = list(entries.values())[0]
        assert_default_test_entry(entry)

    def test_event_post_arxiv_parse(self, caplog: pytest.LogCaptureFixture) -> None:
        """Tests the PostArxivParse event."""

        @Event.PostArxivParse.subscribe
        def hook(bib: Dict[str, Entry]) -> None:
            bib["Cao2018"].data["test"] = "dummy"

        assert Event.PostArxivParse.validate()

        entries = ArxivParser().parse("1812.09976")
        if (
            "cobib.parsers.arxiv",
            logging.ERROR,
            "An Exception occurred while trying to query the arXiv ID: 1812.09976.",
        ) in caplog.record_tuples:
            pytest.skip("The requests API encountered an error. Skipping test.")

        entry = list(entries.values())[0]
        assert_default_test_entry(entry)
        assert entry.data["test"] == "dummy"



"""coBib's terminal user interface.

This class implements a beautiful TUI for coBib, leveraging
[`textual`](https://github.com/textualize/textual).

.. warning::

   This module makes no API stability guarantees! With it being based on
   [`textual`](https://textual.textualize.io/) which is still in very early stages of its
   development, breaking API changes in this module might be released as part of coBib's feature
   releases. You have been warned.
"""

from __future__ import annotations

import asyncio
import io
import logging
import shlex
import sys
from contextlib import contextmanager, redirect_stderr, redirect_stdout
from typing import Any, Awaitable, Callable, Coroutine, Iterator, cast

from rich.console import RenderableType
from textual.app import App, ComposeResult
from textual.css.query import NoMatches
from textual.keys import Keys
from textual.logging import TextualHandler
from textual.widget import AwaitMount, Widget
from textual.widgets import Footer, Header, Input, Static
from typing_extensions import override

from cobib import commands
from cobib.ui.components import (
    EntryView,
    HelpScreen,
    InputScreen,
    ListView,
    MainContent,
    MotionKey,
    Popup,
    PopupLoggingHandler,
    PopupPanel,
    Progress,
    Prompt,
    SearchView,
    SelectionFilter,
)
from cobib.ui.ui import UI
from cobib.utils.file_downloader import FileDownloader


# NOTE: pylint and mypy are unable to understand that the `App` interface actually implements `run`
# pylint: disable=abstract-method
class TUI(UI, App[None]):  # type: ignore[misc]
    """The TUI class.

    This class does not support any extra command-line arguments compared to the base class.
    However, it also extends the `textual.app.App` interface. It can be used with the debugging
    tools of textual but starting it requires you to use the `-c` (a.k.a. `--command`) argument of
    `textual run` since `cobib` is installed as a command-line script.

    In short, here is how you can start the TUI using textual:
    ```
    textual run -c cobib
    ```
    This assumes that you are at the root of the cobib development folder. You can include
    additional command-line arguments after `cobib`, but when doing so you will need to put the
    entire command in quotes, like this:
    ```
    textual run -c "cobib -v"
    ```

    Adding the `--dev` argument to `textual run` will connect it to the debugging console which you
    can start in a separate shell via `textual console`.

    For more information on how to debug a textual app, please refer to
    [their documentation](https://textual.textualize.io/guide/devtools/).
    """

    DEFAULT_CSS = """
        Screen {
            layers: default popup overlay;
            align-vertical: bottom;
        }
    """

    BINDINGS = [
        ("q", "quit", "Quit"),
        ("question_mark", "push_screen('help')", "Help"),
        ("underscore", "toggle_layout", "Layout"),
        ("colon", "prompt(':')", "Prompt"),
        ("v", "select", "Select"),
        ("slash", "prompt('/')", "Search"),
        ("a", "prompt('add ')", "Add"),
        ("d", "delete", "Delete"),
        ("e", "edit", "Edit"),
        ("f", "filter", "Filter"),
        ("i", "prompt('import ')", "Import"),
        ("m", "prompt('modify ', False, True)", "Modify"),
        ("o", "open", "Open"),
        ("r", "prompt('redo', True)", "Redo"),
        ("s", "sort", "Sort"),
        ("u", "prompt('undo', True)", "Undo"),
        ("x", "prompt('export ', False, True)", "Export"),
    ]
    """
    | Key(s) | Description |
    | :- | :- |
    | q | Quit's coBib. |
    | ? | Toggles the help screen. |
    | _ | Toggles between the horizontal and vertical layout. |
    | space | Toggles folding of a search result. |
    | : | Starts the prompt for an arbitrary coBib command. |
    | v | Selects the current entry. |
    | / | Searches the database for the provided string. |
    | a | Prompts for a new entry to be added to the database. |
    | d | Deletes the current (or selected) entries. |
    | e | Edits the current entry. |
    | f | Allows filtering the table using `++/--` keywords. |
    | i | Imports entries from another source. |
    | m | Prompts for a modification (respects selection). |
    | o | Opens the current (or selected) entries. |
    | r | Redoes the last undone change. Requires git-tracking! |
    | s | Prompts for the field to sort by (use -r to list in reverse). |
    | u | Undes the last change. Requires git-tracking! |
    | x | Exports the current (or selected) entries. |
    """

    SCREENS = {
        "help": HelpScreen,
        "input": InputScreen,
    }

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initializes the TUI.

        Args:
            *args: any positional arguments for textual's underlying `App` class.
            **kwargs: any keyword arguments for textual's underlying `App` class.
        """
        super().__init__(*args, **kwargs)
        self.root_logger.addHandler(TextualHandler())
        self.title = "coBib"
        self.sub_title = "The Console Bibliography Manager"
        self._list_args = ["-r"]
        self._filter: SelectionFilter = SelectionFilter()
        self._filters.append(self._filter)
        self._background_tasks: set[asyncio.Task] = set()  # type: ignore[type-arg]
        PopupLoggingHandler(self, level=logging.INFO)
        FileDownloader.console = self
        FileDownloader.progress = Progress

    @override
    def compose(self) -> ComposeResult:
        with MainContent(initial=ListView.id):
            yield ListView()
            yield SearchView(".")

        yield EntryView()

        yield PopupPanel()

        yield Header()
        yield Footer()

    # TODO: remove once https://github.com/Textualize/textual/pull/1541 is merged into Textual
    @contextmanager
    def suspend(self) -> Iterator[None]:
        """Temporarily suspends the application.

        .. warning::

           This method will be removed once support for this has been merged directly into textual.
           Refer to [this pull request](https://github.com/Textualize/textual/pull/1541) for more
           details.
        """
        driver = self._driver
        if driver is not None:
            driver.stop_application_mode()
            with redirect_stdout(sys.__stdout__), redirect_stderr(sys.__stderr__):
                yield
            driver.start_application_mode()

    def print(self, renderable: RenderableType | Widget) -> tuple[Widget, AwaitMount]:
        """A utility method for "printing" to the screen.

        Args:
            renderable: the object to be printed. If this is a `Widget`, it will be mounted in the
                `PopupPanel` as is. Otherwise, the renderable object will be wrapped in a `Popup`
                before mounting without any styling or timer applied.

        Returns:
            The pair of the widget mounted to the `PopupPanel` and the awaitable of its `mount`
            action.
        """
        if isinstance(renderable, Widget):
            popup = renderable
        else:
            popup = Popup(renderable, level=0, timer=None)

        panel = self.screen.query_one(PopupPanel)

        try:
            await_mount = panel.mount(popup, before="#live")
        except NoMatches:
            await_mount = panel.mount(popup)

        return popup, await_mount

    # Event reaction methods

    def on_motion_key(self, event: MotionKey) -> None:
        """Triggers on the custom `cobib.ui.components.motion_key.MotionKey` event.

        This function will update the entry shown in the `cobib.ui.components.entry_view.EntryView`
        widget. It only triggers on vertical motion keys.

        Args:
            event: the motion event.
        """
        if event.key in {Keys.Down, Keys.Up, Keys.PageDown, Keys.PageUp, Keys.Home, Keys.End}:
            self._show_entry()

    async def on_mount(self) -> None:
        """Triggers on the [`Mount`][1] event.

        This method takes care of initializing the desired layout and seeds the `EntryView` widget.

        [1]: https://textual.textualize.io/api/events/#textual.events.Mount
        """
        self.screen.styles.layout = "horizontal"
        await self._update_table()
        self._show_entry()

    # Action methods

    async def action_quit(self) -> None:
        """Action to display the quit dialog."""

        async def _prompt_quit() -> None:
            res = await Prompt.ask(  # type: ignore[call-overload]
                "Are you sure you want to quit?",
                choices=["y", "n"],
                default="y",
                console=self,
            )
            if res == "y":
                self.exit()

        task = asyncio.create_task(_prompt_quit())
        self._background_tasks.add(task)
        task.add_done_callback(self._background_tasks.discard)

    def action_toggle_layout(self) -> None:
        """The layout toggling action.

        This action will toggle between a horizontal and vertical layout.
        """
        layout = self.screen.styles.layout
        if layout is None:
            return
        main = self.query_one(MainContent)
        if layout.name.lower().startswith("horizontal"):
            self.screen.styles.layout = "vertical"
            main.styles.height = "2fr"
            main.styles.width = "1fr"
        elif layout.name.lower().startswith("vertical"):
            self.screen.styles.layout = "horizontal"
            main.styles.width = "2fr"
            main.styles.height = "1fr"
        self.refresh(layout=True)

    async def action_prompt(
        self, value: str, submit: bool = False, check_selection: bool = False
    ) -> None:
        """The prompt action.

        This action triggers an interactive user input via which an arbitrary coBib command can be
        executed as if it were run via the CLI.

        All commands which do not have an explicit action associated with them will go through this
        prompt action.

        Args:
            value: the string with which to pre-populate the input field.
            submit: if set, the user will not be prompted and the input will be submitted
                immediately (i.e. `value` gets parsed directly).
            check_selection: whether or not to check for a visual selection.
        """
        if check_selection and self._filter.selection:
            self._filter.active = True
            value += "-s "

        if submit:
            await self._process_input(value)
        else:
            await_mount = self.push_screen("input", self._process_input)
            inp_screen = cast(InputScreen, self.get_screen("input"))
            inp_screen.escape_enabled = True
            await await_mount
            inp = inp_screen.query_one(Input)
            inp.value = value
            inp.action_end()

    async def action_select(self) -> None:
        """The selection action.

        This action adds the entry currently under the cursor to the visual selection.
        """
        main = self.query_one(MainContent)
        label = main.get_current_label()

        if label in self._filter.selection:
            self._filter.selection.remove(label)
        else:
            self._filter.selection.add(label)

        main.notify_style_update()
        main.refresh()
        self.query_one(EntryView).query_one(Static).refresh()

    def action_delete(self) -> None:
        """The delete action.

        This action triggers the `cobib.commands.delete.DeleteCommand`.
        """
        main = self.query_one(MainContent)
        labels: list[str]
        if self._filter.selection:
            labels = list(self._filter.selection)
            self._filter.selection.clear()
        else:
            labels = [main.get_current_label()]

        self._run_command(["delete"] + labels)

    async def action_edit(self) -> None:
        """The edit action.

        This action triggers the `cobib.commands.edit.EditCommand`.
        """
        self._edit_entry()
        await self._update_table()

    def action_open(self) -> None:
        """The open action.

        This action triggers the `cobib.commands.open.OpenCommand`.
        """
        main = self.query_one(MainContent)
        labels: list[str]
        if self._filter.selection:
            labels = list(self._filter.selection)
            self._filter.selection.clear()
        else:
            labels = [main.get_current_label()]

        self._run_command(["open"] + labels)

    async def action_filter(self) -> None:
        """The filter action.

        This action triggers the `cobib.commands.list.ListCommand` via the `action_prompt` and
        allows the user to provide additional filters.
        """
        await self.action_prompt("list " + " ".join(self._list_args) + " ")

    async def action_sort(self) -> None:
        """The sort action.

        This action triggers the `cobib.commands.list.ListCommand` via the `action_prompt` and
        allows the user to provide a field to be sorted by. This command is handled specially
        because it ensures that only a single sort argument can be given at a time.
        """
        try:
            # first, remove any previously used sort argument
            sort_arg_idx = self._list_args.index("-s")
            # NOTE: we need to pop twice in order to remove both: the `-s` option and the key which
            # was sorted by
            self._list_args.pop(sort_arg_idx)
            self._list_args.pop(sort_arg_idx)
        except ValueError:
            pass

        # add the sort option to the arguments
        self._list_args += ["-s"]

        await self.action_prompt("list " + " ".join(self._list_args) + " ")

    # Utility methods

    def _edit_entry(self) -> None:
        """Edits the current entry.

        This method takes care of suspending the application in favor of the editor.
        """
        main = self.query_one(MainContent)
        label = main.get_current_label()
        with self.suspend():
            commands.EditCommand(label).execute()

    def _show_entry(self) -> None:
        """Renders the entry currently under the cursor in the `EntryView` widget."""
        main = self.query_one(MainContent)
        label = main.get_current_label()
        show_cmd = commands.ShowCommand(label)
        show_cmd.execute()
        entry = self.query_one(EntryView)
        static = entry.query_one(Static)
        static.update(
            show_cmd.render_rich(
                background_color=entry.background_colors[1].rich_color.name,
            )
        )

    async def _update_table(self) -> None:
        """Updates the list of entries displayed in the `MainContent`."""
        # TODO: retain scroll position
        command = commands.ListCommand(*self._list_args)
        command.execute()
        table = command.render_textual()
        main = self.query_one(MainContent)
        await main.replace_widget(table)
        table.focus()
        self.refresh(layout=True)

    async def _update_tree(self, command: list[str]) -> None:
        """Updates the tree of search results displayed in the `MainContent`."""
        subcmd = commands.SearchCommand(*command)
        subcmd.execute()
        tree = subcmd.render_textual()
        main = self.query_one(MainContent)
        await main.replace_widget(tree)
        tree.focus()
        self.refresh(layout=True)

    async def _process_input(self, value: str) -> None:
        """Processes the input returned from the `cobib.ui.components.input_screen.InputScreen`.

        Args:
            value: the value put into the `Input` widget by the user.
        """
        if not value:
            return

        if value[0] == "/":
            value = "search " + value[1:]
        elif value[0] == ":":
            value = value[1:]

        command = shlex.split(value)

        if self._filter.active:
            command += ["--"]
            command.extend(list(self._filter.selection))
            self._filter.active = False

        if command and command[0]:
            if command[0].lower() == "list":
                self._list_args = command[1:]
                await self._update_table()
            elif command[0].lower() == "search":
                await self._update_tree(command[1:])
            else:
                self._run_command(command)

    @staticmethod
    async def _async_done_callback(
        task: Awaitable[None], async_callback: Callable[[], Coroutine[Any, Any, None]]
    ) -> None:
        """Adds an asynchronous callback for when the provided task is done.

        Args:
            task: the Task to await before running the next coroutine.
            async_callback: the asynchronous callback to run after the `task` has completed.
        """
        await task
        await async_callback()

    def _run_command(self, command: list[str]) -> None:
        """Parses and executes a cobib command with its arguments.

        This method also redirects `stdout` and `stderr` and captures their contents to be displayed
        as popups in the TUI.

        Args:
            command: the list of command and its arguments.
        """
        with redirect_stdout(io.StringIO()) as stdout:
            with redirect_stderr(io.StringIO()) as stderr:
                try:
                    subcmd = getattr(commands, command[0].title() + "Command")(
                        *command[1:], prompt=Prompt, console=self
                    )

                    # 1. create subcommand execution task
                    task1 = asyncio.create_task(subcmd.execute())

                    # 2. create another task which chains an asynchronous callback after the
                    #    previous one
                    task2 = asyncio.create_task(TUI._async_done_callback(task1, self._update_table))

                    # 3. ensure proper clean-up of all created tasks
                    self._background_tasks.add(task1)
                    task1.add_done_callback(self._background_tasks.discard)
                    self._background_tasks.add(task2)
                    task2.add_done_callback(self._background_tasks.discard)

                except SystemExit:
                    pass

        stdout_val = stdout.getvalue().strip()
        if stdout_val:
            self.print(Popup(stdout_val, level=logging.INFO))

        stderr_val = stderr.getvalue().strip()
        if stderr_val:
            self.print(Popup(stderr_val, level=logging.CRITICAL))



"""Tests for coBib's DeleteCommand."""
# pylint: disable=unused-argument

from __future__ import annotations

import contextlib
import tempfile
from io import StringIO
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Type

import pytest
from typing_extensions import override

from cobib.commands import DeleteCommand
from cobib.config import Event, config
from cobib.database import Database
from cobib.utils.rel_path import RelPath

from .. import MockStdin, get_resource
from .command_test import CommandTest

if TYPE_CHECKING:
    import _pytest.fixtures

    import cobib.commands


class TestDeleteCommand(CommandTest):
    """Tests for coBib's DeleteCommand."""

    @override
    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        return DeleteCommand

    @pytest.fixture
    def post_setup(
        self, monkeypatch: pytest.MonkeyPatch, request: _pytest.fixtures.SubRequest
    ) -> Generator[Dict[str, Any], None, None]:
        """Additional setup instructions.

        Args:
            monkeypatch: the built-in pytest fixture.
            request: a pytest sub-request providing access to nested parameters.

        Yields:
            The internally used parameters for potential later re-use during the actual test.
        """
        if not hasattr(request, "param"):
            # use default settings
            request.param = {"stdin_list": None}

        monkeypatch.setattr("sys.stdin", MockStdin(request.param.get("stdin_list", None)))

        yield request.param

    def _assert(self, labels: List[str]) -> None:
        """Common assertion utility method.

        Args:
            labels: the list of labels to be deleted.
        """
        bib = Database()

        for label in labels:
            assert bib.get(label, None) is None

        with open(config.database.file, "r", encoding="utf-8") as file:
            with open(get_resource("example_literature.yaml"), "r", encoding="utf-8") as expected:
                # NOTE: do NOT use zip_longest to omit last entries (for testing simplicity)
                for line, truth in zip(file, expected):
                    assert line == truth
                with pytest.raises(StopIteration):
                    next(file)

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "setup",
        [
            {"git": False},
            {"git": True},
        ],
        indirect=["setup"],
    )
    @pytest.mark.parametrize(
        ["labels", "skip_commit"],
        [
            [["knuthwebsite"], False],
            [["knuthwebsite", "latexcompanion"], False],
            # non-existent labels should not cause any problems (but skip git check)
            [["dummy"], True],
            [["dummy", "knuthwebsite"], False],
        ],
    )
    async def test_command(self, setup: Any, labels: List[str], skip_commit: bool) -> None:
        # pylint: disable=invalid-overridden-method
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            labels: the list of labels to be deleted.
            skip_commit: whether to skip asserting the git commit message.
        """
        git = setup.get("git", False)

        # delete some data (for testing simplicity we delete the entries from the end)
        await DeleteCommand(*labels).execute()
        self._assert(labels)

        if git and not skip_commit:
            # assert the git commit message
            self.assert_git_commit_message("delete", {"labels": labels, "preserve_files": None})

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["labels", "post_setup"],
        [
            [["knuthwebsite"], {"stdin_list": ["y"]}],
            [["knuthwebsite"], {"stdin_list": ["n"]}],
            [["knuthwebsite", "latexcompanion"], {"stdin_list": ["y", "y"]}],
            [["knuthwebsite", "latexcompanion"], {"stdin_list": ["n", "n"]}],
            [["knuthwebsite", "latexcompanion"], {"stdin_list": ["y", "n"]}],
            [["knuthwebsite", "latexcompanion"], {"stdin_list": ["n", "y"]}],
        ],
        indirect=["post_setup"],
    )
    async def test_confirmation_prompt(
        self, setup: Any, post_setup: Any, labels: List[str]
    ) -> None:
        """TODO."""
        config.commands.delete.confirm = True

        choices = post_setup.get("stdin_list", False)

        await DeleteCommand(*labels).execute()

        if "n" in choices:
            bib = Database()

            for label, choice in zip(labels, choices):
                entry = bib.get(label, None)
                if choice == "y":
                    assert entry is None
                else:
                    assert entry is not None
        else:
            self._assert(labels)

    @pytest.mark.asyncio
    @pytest.mark.parametrize("preserve_files", [None, True, False])
    @pytest.mark.parametrize("config_overwrite", [True, False])
    async def test_remove_associated_file(
        self, setup: Any, preserve_files: bool, config_overwrite: bool
    ) -> None:
        """Test removing associated files.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            preserve_files: argument to `DeleteCommand`.
            config_overwrite: what to overwrite `config.commands.delete.preserve_files` with.
        """
        config.commands.delete.preserve_files = config_overwrite

        should_preserve = config_overwrite
        if preserve_files is not None:
            should_preserve = preserve_files

        with tempfile.TemporaryDirectory() as tmpdirname:
            path = RelPath(tmpdirname + "/dummy.pdf")
            open(path.path, "w", encoding="utf-8").close()  # pylint: disable=consider-using-with

            Database()["knuthwebsite"].file = str(path)

            args = ["knuthwebsite"]
            if preserve_files is not None:
                args += [f"--{'' if preserve_files else 'no-'}preserve-files"]
            await DeleteCommand(*args).execute()

            assert path.path.exists() is should_preserve

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        "setup",
        [{"git": False}],
        indirect=["setup"],
    )
    async def test_base_cmd_insufficient_git(
        self, setup: Any, caplog: pytest.LogCaptureFixture
    ) -> None:
        """Test warning is raised by BaseCommand during insufficient git-configuration.

        While this is technically not related to the DeleteCommand, this is one of the faster
        commands to test this on.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            caplog: the built-in pytest fixture.
        """
        config.database.git = True

        await DeleteCommand("knuthwebsite").execute()
        self._assert(["knuthwebsite"])

        assert (
            "cobib.commands.base_command",
            30,
            "You have configured coBib to track your database with git."
            "\nPlease run `cobib init --git`, to initialize this tracking.",
        ) in caplog.record_tuples

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["labels"],
        [
            [["knuthwebsite"]],
            [["knuthwebsite", "latexcompanion"]],
        ],
    )
    # other variants are already covered by test_command
    async def test_cmdline(
        self, setup: Any, monkeypatch: pytest.MonkeyPatch, labels: List[str]
    ) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
            labels: the list of labels to be deleted.
        """
        await self.run_module(monkeypatch, "main", ["cobib", "delete"] + labels)
        self._assert(labels)

    @pytest.mark.asyncio
    async def test_event_pre_delete_command(self, setup: Any) -> None:
        """Tests the PreDeleteCommand event."""

        @Event.PreDeleteCommand.subscribe
        def hook(command: DeleteCommand) -> None:
            command.largs.labels = ["einstein"]

        assert Event.PreDeleteCommand.validate()

        await DeleteCommand("knuthwebsite").execute()

        assert "einstein" not in Database().keys()
        assert "knuthwebsite" in Database().keys()

    @pytest.mark.asyncio
    async def test_event_post_delete_command(self, setup: Any) -> None:
        """Tests the PostDeleteCommand event."""

        @Event.PostDeleteCommand.subscribe
        def hook(command: DeleteCommand) -> None:
            for label in command.deleted_entries:
                print(f"WARNING: deleted entry '{label}'")

        with contextlib.redirect_stdout(StringIO()) as out:
            await DeleteCommand("knuthwebsite").execute()

        assert Event.PostDeleteCommand.validate()

        assert out.getvalue().strip() == "WARNING: deleted entry 'knuthwebsite'"



"""coBib importer test class."""

import pytest

from cobib.config import config


class ImporterTest:
    """The base class for coBib's importer test classes."""

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        """Setup."""
        config.defaults()



"""Tests for coBib's SearchCommand."""
# pylint: disable=unused-argument

from __future__ import annotations

import contextlib
from io import StringIO
from itertools import zip_longest
from typing import TYPE_CHECKING, Any, List, Type

import pytest
from rich.console import Console
from rich.tree import Tree
from typing_extensions import override

from cobib.commands import SearchCommand
from cobib.config import Event, config

from .command_test import CommandTest

if TYPE_CHECKING:
    import cobib.commands


class TestSearchCommand(CommandTest):
    """Tests for coBib's SearchCommand."""

    @override
    def get_command(self) -> Type[cobib.commands.base_command.Command]:
        return SearchCommand

    def _assert(self, output: List[str], expected: List[str]) -> None:
        """Common assertion utility method.

        Args:
            output: the list of lines printed to `sys.stdout`.
            expected: the expected output.
        """
        # we use zip_longest to ensure that we don't have more than we expect
        for line, exp in zip_longest(output, expected):
            assert line == exp

    @pytest.mark.parametrize(
        ["args", "expected", "config_overwrite"],
        [
            [
                ["einstein"],
                ["einstein::1", "1::@article{einstein,", "1::author = {Albert Einstein},"],
                False,
            ],
            [
                ["einstein", "--", "--label", "einstein"],
                [],
                False,
            ],
            [
                ["einstein", "-i"],
                [
                    "einstein::2",
                    "1::@article{einstein,",
                    "2::author = {Albert Einstein},",
                    "2::doi = {http://dx.doi.org/10.1002/andp.19053221004},",
                ],
                False,
            ],
            [
                ["einstein", "-i", "-c", "0"],
                ["einstein::2", "1::@article{einstein,", "2::author = {Albert Einstein},"],
                False,
            ],
            [
                ["einstein", "-i", "-c", "2"],
                [
                    "einstein::2",
                    "1::@article{einstein,",
                    "2::author = {Albert Einstein},",
                    "2::doi = {http://dx.doi.org/10.1002/andp.19053221004},",
                    "2::journal = {Annalen der Physik},",
                ],
                False,
            ],
            [
                ["einstein"],
                [
                    "einstein::2",
                    "1::@article{einstein,",
                    "2::author = {Albert Einstein},",
                    "2::doi = {http://dx.doi.org/10.1002/andp.19053221004},",
                ],
                True,
            ],
            [
                ["einstein", "-i"],
                [
                    "einstein::2",
                    "1::@article{einstein,",
                    "2::author = {Albert Einstein},",
                    "2::doi = {http://dx.doi.org/10.1002/andp.19053221004},",
                ],
                True,
            ],
            [
                ["einstein", "Elektro"],
                [
                    "einstein::2",
                    "1::@article{einstein,",
                    "1::author = {Albert Einstein},",
                    "2::pages = {891--921},",
                    r"2::title = {Zur Elektrodynamik bewegter K{\"o}rper},",
                    "2::volume = {322},",
                ],
                False,
            ],
            [
                ["einstein", "Elektro", "--", "--label", "einstein"],
                [],
                False,
            ],
            [
                ["einstein", "-I"],
                ["einstein::1", "1::@article{einstein,", "1::author = {Albert Einstein},"],
                True,
            ],
        ],
    )
    def test_command(
        self, setup: Any, args: List[str], expected: List[str], config_overwrite: bool
    ) -> None:
        """Test the command itself.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            args: the arguments to pass to the command.
            expected: the expected output.
            config_overwrite: what to overwrite `config.commands.search.ignore_case` with.
        """
        config.commands.search.ignore_case = config_overwrite

        cmd = SearchCommand(*args)
        cmd.execute()
        output = cmd.render_porcelain()
        self._assert(output, expected)

    def test_context_configuration(self, setup: Any) -> None:
        """Test the `config.commands.search.context` setting.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """
        config.commands.search.context = 2

        cmd = SearchCommand(
            "einstein",
            "-i",
            "-c",
            "2",
        )
        cmd.execute()
        output = cmd.render_porcelain()
        self._assert(
            output,
            [
                "einstein::2",
                "1::@article{einstein,",
                "2::author = {Albert Einstein},",
                "2::doi = {http://dx.doi.org/10.1002/andp.19053221004},",
                "2::journal = {Annalen der Physik},",
            ],
        )

    def test_render_rich(self, setup: Any) -> None:
        """Test the rich rendering.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
        """
        cmd = SearchCommand("einstein", "-i")
        cmd.execute()
        renderable = cmd.render_rich()

        assert isinstance(renderable, Tree)
        console = Console(record=True)
        console.print(renderable)
        assert console.export_text() == (
            "einstein - 2 matches\n"
            "├── 1\n"
            "│   └── @article{einstein,\n"
            "└── 2\n"
            "    ├──  author = {Albert Einstein},\n"
            "    └──  doi = {http://dx.doi.org/10.1002/andp.19053221004},\n"
        )

    @pytest.mark.asyncio
    @pytest.mark.parametrize(
        ["expected"],
        [
            [["einstein::1", "1::@article{einstein,", "1::author = {Albert Einstein},"]],
        ],
    )
    # other variants are already covered by test_command
    async def test_cmdline(
        self,
        setup: Any,
        monkeypatch: pytest.MonkeyPatch,
        capsys: pytest.CaptureFixture[str],
        expected: List[str],
    ) -> None:
        """Test the command-line access of the command.

        Args:
            setup: the `tests.commands.command_test.CommandTest.setup` fixture.
            monkeypatch: the built-in pytest fixture.
            capsys: the built-in pytest fixture.
            expected: the expected output.
        """
        await self.run_module(monkeypatch, "main", ["cobib", "--porcelain", "search", "einstein"])
        self._assert(capsys.readouterr().out.strip().split("\n"), expected)

    def test_event_pre_search_command(self, setup: Any) -> None:
        """Tests the PreSearchCommand event."""

        @Event.PreSearchCommand.subscribe
        def hook(command: SearchCommand) -> None:
            command.largs.query = ["einstein"]

        assert Event.PreSearchCommand.validate()

        expected = ["einstein::1", "1::@article{einstein,", "1::author = {Albert Einstein},"]

        cmd = SearchCommand("knuthwebsite")
        cmd.execute()
        output = cmd.render_porcelain()
        self._assert(output, expected)

    def test_event_post_search_command(self, setup: Any) -> None:
        """Tests the PostSearchCommand event."""

        @Event.PostSearchCommand.subscribe
        def hook(command: SearchCommand) -> None:
            print(command.hits, [entry.label for entry in command.entries])

        assert Event.PostSearchCommand.validate()

        with contextlib.redirect_stdout(StringIO()) as file:
            cmd = SearchCommand("einstein")
            cmd.execute()
            assert file.getvalue().strip() == "1 ['einstein']"



"""coBib's subscribable events.

### Available Events

There are various kinds of event types:
    - Pre*Command: these fire before a command gets executed. Hooks subscribing to these events
      are passed an instance of the command which will be populated with the original command-line
      arguments as well as the resulting `argparse.Namespace`.
    - Post*Command: these fire after a command got executed but (generally) **before** the
      Database gets written to file, allowing final touch-ups and modifications to take place.
      Just like the `Pre*Command` events, the input will be an instance of the command through which
      a user can modify the command data at runtime.
    - Pre*Import: these fire before an importer gets executed. Hooks subscribing to these events
      are passed an instance of the importer which will be populated with the original command-line
      arguments as well as the resulting `argparse.Namespace`.
    - Post*Import: these fire after an importer got executed but (generally) **before** the
      Database gets written to file, allowing final touch-ups and modifications to take place.
      Just like the `Pre*Import` events, the input will be an instance of the importer through which
      a user can modify the command data at runtime.
    - Pre*Parse: just like the Pre-Command events, these fire before a parser runs. As an input
      they generally get the driver input.
    - Post*Parse: these fire after a parser ran. They again allow final touch-ups of the
      generated dictionary of new entries.
    - Pre*Dump: these fire before an entry gets dumped. The hook can pre-process the entry to
      its desire. Changes will not become persistent in the Database.
    - Post*Dump: these fire after an entry got formatted as a string. The string can be
      post-processed with some final touch-ups.
    - and finally there a few specific events not belonging to any of the categories mentioned
      above, examples of which are the `PreGitCommit` and `PostGitCommit` events.

All events are listed below.

### Usage

You can register a function to be executed when a certain event gets triggered as shown in the
following example:
```python
from os import system
from cobib.config import Event
from cobib.commands import InitCommand

@Event.PostInitCommand.subscribe
def add_remote(cmd: InitCommand) -> None:
    system(f"git -C {cmd.root} remote add origin https://github.com/user/repo")
```
The above example gets run after the `init` command has finished. It adds a remote to the git
repository. This can be useful in combination with automatic pushing to the remote like done here:
```python
from pathlib import Path
from os import system
from cobib.config import Event

@Event.PostGitCommit.subscribe
def push_to_remote(root: Path, file: Path) -> None:
    system(f"git -C {root} push origin master")
```

It is important that you include the type hints as part of the function definition because these are
used during the config validation.

You can also subscribe the same function to multiple events at once as shown in the following
(rather non-sensible) example:
```python
from typing import Dict
from cobib.database import Entry

@Event.PostArxivParse.subscribe
@Event.PostDOIParse.subscribe
@Event.PostISBNParse.subscribe
def print_new_entries(bib: Dict[str, Entry) -> None:
    print("New entries being added: ", list(bib.keys()))
```
"""

from __future__ import annotations

import logging
from enum import Enum
from itertools import zip_longest
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    ForwardRef,
    List,
    Optional,
    Tuple,
    get_type_hints,
)

from cobib.utils.rel_path import RelPath

from .config import config

_FORWARD_REFS: Dict[str, str] = {}
if TYPE_CHECKING:
    from cobib import commands, importers
    from cobib.database import Entry
else:
    _FORWARD_REFS = {
        "ForwardRef('Entry')": "Entry",
        "ForwardRef('commands.AddCommand')": "cobib.commands.add.AddCommand",
        "ForwardRef('commands.DeleteCommand')": "cobib.commands.delete.DeleteCommand",
        "ForwardRef('commands.EditCommand')": "cobib.commands.edit.EditCommand",
        "ForwardRef('commands.ExportCommand')": "cobib.commands.export.ExportCommand",
        "ForwardRef('commands.ImportCommand')": "cobib.commands.import_.ImportCommand",
        "ForwardRef('commands.InitCommand')": "cobib.commands.init.InitCommand",
        "ForwardRef('commands.ListCommand')": "cobib.commands.list_.ListCommand",
        "ForwardRef('commands.ModifyCommand')": "cobib.commands.modify.ModifyCommand",
        "ForwardRef('commands.OpenCommand')": "cobib.commands.open.OpenCommand",
        "ForwardRef('commands.RedoCommand')": "cobib.commands.redo.RedoCommand",
        "ForwardRef('commands.SearchCommand')": "cobib.commands.search.SearchCommand",
        "ForwardRef('commands.ShowCommand')": "cobib.commands.show.ShowCommand",
        "ForwardRef('commands.UndoCommand')": "cobib.commands.undo.UndoCommand",
        "ForwardRef('importers.ZoteroImporter')": "cobib.importers.zotero.ZoteroImporter",
    }

LOGGER = logging.getLogger(__name__)


class Event(Enum):
    # pylint: disable=invalid-name
    """Subscribable events triggered at runtime.

    The following sections list all events available in coBib. The heading of each section includes
    the name of the event, the type (always `Event`) and its unique identifier which is a tuple made
    up of a unique integer ID plus the `Callable` type hint required for the hooks subscribing to
    this particular event.
    """

    _annotation_: Any

    def __new__(cls, annotation: Any) -> Event:
        """Enum constructor.

        We overwrite the default constructor to handle the "overloaded" use of the Enum values in
        which we store the non-unique Callable type hints. The actual values of the Enum are simply
        integers.

        Args:
            annotation: the expected Callable type hint for the hooks registering to this event.

        Returns:
            The new Event instance.
        """
        value = len(cls.__members__) + 1
        obj = object.__new__(cls)
        obj._value_ = (value, annotation)
        obj._annotation_ = annotation
        return obj

    PreAddCommand: Event = Callable[["commands.AddCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before starting the `cobib.commands.add.AddCommand`.

    Arguments:
        - `cobib.commands.add.AddCommand`: the command instance that is about to run.

    Returns:
        Nothing. But the command attributes can be modified, affecting the execution.
    """
    PostAddCommand: Event = Callable[["commands.AddCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before finishing the `cobib.commands.add.AddCommand`.

    Arguments:
        - `cobib.commands.add.AddCommand`: the command instance that just ran.

    Returns:
        Nothing. But the command attributes can be modified, affecting the execution.

    Note:
        This event fires *before* starting the `cobib.commands.edit.EditCommand` which starts if
        manual entry addition is requested.
    """

    PreDeleteCommand: Event = Callable[["commands.DeleteCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before starting the `cobib.commands.delete.DeleteCommand`.

    Arguments:
        - `cobib.commands.delete.DeleteCommand`: the command instance that is about to run.

    Returns:
        Nothing. But the command attributes can be modified, affecting the execution.
    """
    PostDeleteCommand: Event = Callable[  # type: ignore[assignment]
        ["commands.DeleteCommand"], None
    ]
    """
    Fires:
        Before finishing the `cobib.commands.delete.DeleteCommand`.

    Arguments:
        - `cobib.commands.delete.DeleteCommand`: the command instance that just ran.

    Returns:
        Nothing. While the deleted entry labels are accessible, modifying them has no effect.
    """

    PreEditCommand: Event = Callable[["commands.EditCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before starting the `cobib.commands.edit.EditCommand`.

    Arguments:
        - `cobib.commands.edit.EditCommand`: the command instance that is about to run.

    Returns:
        Nothing. But the command attributes can be modified, affecting the execution.
    """
    PostEditCommand: Event = Callable[["commands.EditCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before finishing the `cobib.commands.edit.EditCommand`.

    Arguments:
        - `cobib.commands.edit.EditCommand`: the command instance that just ran.

    Returns:
        Nothing. While the edited entry is accessible, modifying it has no effect.
    """

    PreExportCommand: Event = Callable[["commands.ExportCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before starting the `cobib.commands.export.ExportCommand`.

    Arguments:
        - `cobib.commands.export.ExportCommand`: the command instance that is about to run.

    Returns:
        Nothing. But the command attributes can be modified, affecting the execution.
    """
    PostExportCommand: Event = Callable[  # type: ignore[assignment]
        ["commands.ExportCommand"], None
    ]
    """
    Fires:
        Before finishing the `cobib.commands.export.ExportCommand`.

    Arguments:
        - `cobib.commands.export.ExportCommand`: the command instance that just ran.

    Returns:
        Nothing. The files to which has been exported are still accessible and open.
    """

    PreImportCommand: Event = Callable[["commands.ImportCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before starting the `cobib.commands.import_.ImportCommand`.

    Arguments:
        - `cobib.commands.import_.ImportCommand`: the command instance that is about to run.

    Returns:
        Nothing. But the command attributes can be modified, affecting the execution.
    """
    PostImportCommand: Event = Callable[  # type: ignore[assignment]
        ["commands.ImportCommand"], None
    ]
    """
    Fires:
        Before finishing the `cobib.commands.import_.ImportCommand`.

    Arguments:
        - `cobib.commands.import_.ImportCommand`: the command instance that just ran.

    Returns:
        Nothing. But the dictionary of new entries can be modified before the changes are made
        persistent in the database.
    """

    PreInitCommand: Event = Callable[["commands.InitCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before starting the `cobib.commands.init.InitCommand`.

    Arguments:
        - `cobib.commands.init.InitCommand`: the command instance that is about to run.

    Returns:
        Nothing. But the command attributes can be modified, affecting the execution.
    """
    PostInitCommand: Event = Callable[["commands.InitCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before finishing the `cobib.commands.init.InitCommand`.

    Arguments:
        - `cobib.commands.init.InitCommand`: the command instance that just ran.

    Returns:
        Nothing.
    """

    PreListCommand: Event = Callable[["commands.ListCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before starting the `cobib.commands.list_.ListCommand`.

    Arguments:
        - `cobib.commands.list_.ListCommand`: the command instance that is about to run.

    Returns:
        Nothing. But the command attributes can be modified, affecting the execution.
    """
    PostListCommand: Event = Callable[["commands.ListCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before finishing the `cobib.commands.list_.ListCommand`.

    Arguments:
        - `cobib.commands.list_.ListCommand`: the command instance that just ran.

    Returns:
        Nothing.
    """

    PreModifyCommand: Event = Callable[["commands.ModifyCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before starting the `cobib.commands.modify.ModifyCommand`.

    Arguments:
        - `cobib.commands.modify.ModifyCommand`: the command instance that is about to run.

    Returns:
        Nothing. But the command attributes can be modified, affecting the execution.
    """
    PostModifyCommand: Event = Callable[  # type: ignore[assignment]
        ["commands.ModifyCommand"], None
    ]
    """
    Fires:
        Before finishing the `cobib.commands.modify.ModifyCommand`.

    Arguments:
        - `cobib.commands.modify.ModifyCommand`: the command instance that just ran.

    Returns:
        Nothing. But the modified entries are still accessible before written to the database.
    """

    PreOpenCommand: Event = Callable[["commands.OpenCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before starting the `cobib.commands.open.OpenCommand`.

    Arguments:
        - `cobib.commands.open.OpenCommand`: the command instance that is about to run.

    Returns:
        Nothing. But the command attributes can be modified, affecting the execution.
    """
    PostOpenCommand: Event = Callable[["commands.OpenCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before finishing the `cobib.commands.open.OpenCommand`.

    Arguments:
        - `cobib.commands.open.OpenCommand`: the command instance that just ran.

    Returns:
        Nothing.
    """

    PreRedoCommand: Event = Callable[["commands.RedoCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before starting the `cobib.commands.redo.RedoCommand`.

    Arguments:
        - `cobib.commands.redo.RedoCommand`: the command instance that is about to run.

    Returns:
        Nothing. But the command attributes can be modified, affecting the execution.
    """
    PostRedoCommand: Event = Callable[["commands.RedoCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before finishing the `cobib.commands.redo.RedoCommand`.

    Arguments:
        - `cobib.commands.redo.RedoCommand`: the command instance that just ran.

    Returns:
        Nothing.
    """

    PreSearchCommand: Event = Callable[["commands.SearchCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before starting the `cobib.commands.search.SearchCommand`.

    Arguments:
        - `cobib.commands.search.SearchCommand`: the command instance that is about to run.

    Returns:
        Nothing. But the command attributes can be modified, affecting the execution.
    """
    PostSearchCommand: Event = Callable[  # type: ignore[assignment]
        ["commands.SearchCommand"], None
    ]
    """
    Fires:
        Before finishing the `cobib.commands.search.SearchCommand`.

    Arguments:
        - `cobib.commands.search.SearchCommand`: the command instance that just ran.

    Returns:
        Nothing. But the search results are still accessible before being rendered for the user.
    """

    PreShowCommand: Event = Callable[["commands.ShowCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before starting the `cobib.commands.show.ShowCommand`.

    Arguments:
        - `cobib.commands.show.ShowCommand`: the command instance that is about to run.

    Returns:
        Nothing. But the command attributes can be modified, affecting the execution.
    """
    PostShowCommand: Event = Callable[["commands.ShowCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before finishing the `cobib.commands.show.ShowCommand`.

    Arguments:
        - `cobib.commands.show.ShowCommand`: the command instance that just ran.

    Returns:
        Nothing. But the string-represented entry is still accessible before being rendered.
    """

    PreUndoCommand: Event = Callable[["commands.UndoCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before starting the `cobib.commands.undo.UndoCommand`.

    Arguments:
        - `cobib.commands.undo.UndoCommand`: the command instance that is about to run.

    Returns:
        Nothing. But the command attributes can be modified, affecting the execution.
    """
    PostUndoCommand: Event = Callable[["commands.UndoCommand"], None]  # type: ignore[assignment]
    """
    Fires:
        Before finishing the `cobib.commands.undo.UndoCommand`.

    Arguments:
        - `cobib.commands.undo.UndoCommand`: the command instance that just ran.

    Returns:
        Nothing.
    """

    PreBibtexParse: Event = Callable[[str], Optional[str]]  # type: ignore[assignment]
    """
    Fires:
        Before starting `cobib.parsers.bibtex.BibtexParser.parse`.

    Arguments:
        - `string`: the string to be parsed as BibTeX.

    Returns:
        Optionally a new (or updated) string to be parsed as BibTeX.

    Note:
        If a registered hook returns a new string, no subsequent hooks will be run!
    """
    PostBibtexParse: Event = Callable[[Dict[str, "Entry"]], None]  # type: ignore[assignment]
    """
    Fires:
        Before finishing `cobib.parsers.bibtex.BibtexParser.parse`.

    Arguments:
        - `bib`: a dictionary of the new `Entry` instances stored under their `label` as keys.

    Returns:
        Nothing. But the dictionary can be modified in-place such that the changes will be
        propagated to the database.
    """
    PreBibtexDump: Event = Callable[["Entry"], None]  # type: ignore[assignment]
    """
    Fires:
        Before starting `cobib.parsers.bibtex.BibtexParser.dump`.

    Arguments:
        - `cobib.database.Entry`: the `Entry` object to be dumped in BibTeX format.

    Returns:
        Nothing. But the object can be modified in-place. Changes will *not* become persistent in
        the database.
    """
    PostBibtexDump: Event = Callable[[str], Optional[str]]  # type: ignore[assignment]
    """
    Fires:
        Before finishing `cobib.parsers.bibtex.BibtexParser.dump`.

    Arguments:
        - `string`: the string-representation of the `Entry` to be dumped as BibTeX.

    Returns:
        Optionally a new (or updated) string to be dumped.

    Note:
        If a registered hook returns a new string, no subsequent hooks will be run!
    """

    PreYAMLParse: Event = Callable[[str], Optional[str]]  # type: ignore[assignment]
    """
    Fires:
        Before starting `cobib.parsers.yaml.YAMLParser.parse`.

    Arguments:
        - `string`: the string to be parsed as YAML.

    Returns:
        Optionally a new (or updated) string to be parsed as YAML.

    Note:
        If a registered hook returns a new string, no subsequent hooks will be run!
    """
    PostYAMLParse: Event = Callable[[Dict[str, "Entry"]], None]  # type: ignore[assignment]
    """
    Fires:
        Before finishing `cobib.parsers.yaml.YAMLParser.parse`.

    Arguments:
        - `bib`: a dictionary of the new `Entry` instances stored under their `label` as keys.

    Returns:
        Nothing. But the dictionary can be modified in-place such that the changes will be
        propagated to the database.
    """
    PreYAMLDump: Event = Callable[["Entry"], None]  # type: ignore[assignment]
    """
    Fires:
        Before starting `cobib.parsers.yaml.YAMLParser.dump`.

    Arguments:
        - `cobib.database.Entry`: the `Entry` object to be dumped in YAML format.

    Returns:
        Nothing. But the object can be modified in-place. Changes will *not* become persistent in
        the database.
    """
    PostYAMLDump: Event = Callable[[str], Optional[str]]  # type: ignore[assignment]
    """
    Fires:
        Before finishing `cobib.parsers.yaml.YAMLParser.dump`.

    Arguments:
        - `string`: the string-representation of the `Entry` to be dumped as YAML.

    Returns:
        Optionally a new (or updated) string to be dumped.

    Note:
        If a registered hook returns a new string, no subsequent hooks will be run!
    """

    PreArxivParse: Event = Callable[[str], Optional[str]]  # type: ignore[assignment]
    """
    Fires:
        Before starting `cobib.parsers.arxiv.ArxivParser.parse`.

    Arguments:
        - `string`: the string to be parsed as an arXiv ID.

    Returns:
        Optionally a new (or updated) string to be parsed as an arXiv ID.

    Note:
        If a registered hook returns a new string, no subsequent hooks will be run!
    """
    PostArxivParse: Event = Callable[[Dict[str, "Entry"]], None]  # type: ignore[assignment]
    """
    Fires:
        Before finishing `cobib.parsers.arxiv.ArxivParser.parse`.

    Arguments:
        - `bib`: a dictionary of the new `Entry` instances stored under their `label` as keys.

    Returns:
        Nothing. But the dictionary can be modified in-place such that the changes will be
        propagated to the database.
    """

    PreDOIParse: Event = Callable[[str], Optional[str]]  # type: ignore[assignment]
    """
    Fires:
        Before starting `cobib.parsers.doi.DOIParser.parse`.

    Arguments:
        - `string`: the string to be parsed as a DOI.

    Returns:
        Optionally a new (or updated) string to be parsed as a DOI.

    Note:
        If a registered hook returns a new string, no subsequent hooks will be run!
    """
    PostDOIParse: Event = Callable[[Dict[str, "Entry"]], None]  # type: ignore[assignment]
    """
    Fires:
        Before finishing `cobib.parsers.doi.DOIParser.parse`.

    Arguments:
        - `bib`: a dictionary of the new `Entry` instances stored under their `label` as keys.

    Returns:
        Nothing. But the dictionary can be modified in-place such that the changes will be
        propagated to the database.
    """

    PreISBNParse: Event = Callable[[str], Optional[str]]  # type: ignore[assignment]
    """
    Fires:
        Before starting `cobib.parsers.isbn.ISBNParser.parse`.

    Arguments:
        - `string`: the string to be parsed as an ISBN.

    Returns:
        Optionally a new (or updated) string to be parsed as an ISBN.

    Note:
        If a registered hook returns a new string, no subsequent hooks will be run!
    """
    PostISBNParse: Event = Callable[[Dict[str, "Entry"]], None]  # type: ignore[assignment]
    """
    Fires:
        Before finishing `cobib.parsers.isbn.ISBNParser.parse`.

    Arguments:
        - `bib`: a dictionary of the new `Entry` instances stored under their `label` as keys.

    Returns:
        Nothing. But the dictionary can be modified in-place such that the changes will be
        propagated to the database.
    """

    PreURLParse: Event = Callable[[str], Optional[str]]  # type: ignore[assignment]
    """
    Fires:
        Before starting `cobib.parsers.url.URLParser.parse`.

    Arguments:
        - `string`: the string to be parsed as a URL.

    Returns:
        Optionally a new (or updated) string to be parsed as a URL.

    Note:
        If a registered hook returns a new string, no subsequent hooks will be run!
    """
    PostURLParse: Event = Callable[[Dict[str, "Entry"]], None]  # type: ignore[assignment]
    """
    Fires:
        Before finishing `cobib.parsers.url.URLParser.parse`.

    Arguments:
        - `bib`: a dictionary of the new `Entry` instances stored under their `label` as keys.

    Returns:
        Nothing. But the dictionary can be modified in-place such that the changes will be
        propagated to the database.
    """

    PreZoteroImport: Event = Callable[  # type: ignore[assignment]
        ["importers.ZoteroImporter"], None
    ]
    """
    Fires:
        Before starting `cobib.importers.zotero.ZoteroImporter.fetch`.

    Arguments:
        - `cobib.importers.zotero.ZoteroImporter`: the importer instance that is about to run.

    Returns:
        Nothing. But the importer attributes can be modified, affecting the execution.
    """
    PostZoteroImport: Event = Callable[  # type: ignore[assignment]
        ["importers.ZoteroImporter"], None
    ]
    """
    Fires:
        Before finishing `cobib.importers.zotero.ZoteroImporter.fetch`.

    Arguments:
        - `cobib.importers.zotero.ZoteroImporter`: the importer instance that is about to run.

    Returns:
        Nothing. But the importer attributes can be modified, affecting the execution.

    Note:
        - The entry labels will not have been mapped or disambiguated at this point.
    """

    PreFileDownload: Event = Callable[  # type: ignore[assignment]
        [str, str, Optional[str], Optional[Dict[str, str]]],
        Optional[Tuple[str, str, Optional[str], Optional[Dict[str, str]]]],
    ]
    """
    Fires:
        Before starting `cobib.utils.file_downloader.FileDownloader.download`.

    Arguments:
        - `url`: the URL from which to download a file.
        - `label`: the label of the `Entry` to which the file belongs.
        - `folder`: an optional folder where the file will be stored.
        - `headers`: an optional headers dictionary for the download `GET` request.

    Returns:
        This can optionally return a tuple overwriting the input arguments.

    Note:
        If a registered hook returns a new tuple of arguments, no subsequent hooks will be run!
    """
    PostFileDownload: Event = Callable[[RelPath], Optional[RelPath]]  # type: ignore[assignment]
    """
    Fires:
        Before finishing `cobib.utils.file_downloader.FileDownloader.download` if and only if the
        download was successful.

    Arguments:
        - `path`: the `RelPath` to the freshly downloaded file.

    Returns:
        An optional new `RelPath`.

    Note:
        If a registered hook returns a new path, no subsequent hooks will be run!
    """

    PreGitCommit: Event = Callable[  # type: ignore[assignment]
        [str, Optional[Dict[str, Any]]], Optional[str]
    ]
    """
    Fires:
        Before starting `cobib.commands.base_command.Command.git` (i.e. whenever an automatic
        git-commit occurs).

    Arguments:
        - `msg`: the commit message.
        - `args`: an optional dictionary of keyword arguments provided to the command which
                  triggered this commit.

    Returns:
        Optionally a new commit message.

    Note:
        If a registered hook returns a new commit message, no subsequent hooks will be run!
    """
    PostGitCommit: Event = Callable[[Path, Path], None]  # type: ignore[assignment]
    """
    Fires:
        Before finishing `cobib.commands.base_command.Command.git` (i.e. whenever an automatic
        git-commit occurs).

    Arguments:
        - `root`: the `Path` to the root git directory where the database file resides.
        - `file`: the `Path` to the database file.

    Returns:
        Nothing.
    """

    def subscribe(self, function: Callable) -> Callable:  # type: ignore[type-arg]
        """Subscribes the provided function to this Event.

        Usage:
            ```python
            @Event.PostGitCommit.subscribe
            def push_to_remote(root: Path, file: Path) -> None:
                os.system(f"git -C {root} push")
            ```

        Args:
            function: the callable to execute when this Event fires.

        Returns:
            The same callable which ensures that we can chain multiple subscription decorators to
            subscribe the same function to multiple events.
        """
        if self not in config.events:
            config.events[self] = []
        LOGGER.debug("Subscribing new hook to %s.", str(self))
        config.events[self].append(function)
        return function

    def fire(self, *args: Any, **kwargs: Any) -> Any:
        """Fires the Event by executing all subscribed hooks.

        Note, that if a hook has a non-None return value this method will return prematurely
        regardless of whether more hooks are subscribed.
        The exact signature of the hooks depends on the Event as indicated above.

        Args:
            args: positional hook arguments.
            kwargs: keyword hook arguments.

        Returns:
            Whatever a hook returns or `None` as a default.
        """
        LOGGER.debug("Firing %s.", str(self))
        if self not in config.events:
            return None
        for hook in config.events[self]:
            res = hook(*args, **kwargs)
            if res is not None:
                return res
        return None

    def validate(self) -> bool:
        """Validates the type hints of all subscribed functions against the expected one.

        Returns:
            Whether or not all subscribers have the correct type hint.
        """
        if self not in config.events:
            return True

        def _compare_types(expected: Any, provided: Any) -> List[bool]:
            result = []
            for exp, pro in zip_longest(expected, provided):
                try:
                    result.extend(_compare_types(exp.__args__, pro.__args__))
                except AttributeError:
                    if isinstance(exp, ForwardRef):
                        result.append(_FORWARD_REFS[str(exp)] in str(pro))
                    elif pro == "None":
                        result.append(exp == type(None))
                    else:
                        result.append(exp == pro)
            return result

        LOGGER.debug("Validating hooks subscribed to %s.", str(self))
        annotation = self._annotation_.__args__
        for idx, sub in enumerate(config.events[self]):
            sub_ann = get_type_hints(sub).values()
            if not all(_compare_types(annotation, sub_ann)):
                LOGGER.debug(
                    "Hook #%s has a mismatching annotation: %s instead of %s",
                    idx,
                    annotation,
                    sub_ann,
                )
                return False
        return True



"""coBib's main content view.

This is a simple subclass of textual's `ContentSwitcher` widget. Any widget mounted into it is
expected to have a `get_current_label` method implement (although this is currently not
programmatically enforced). For examples refer to `cobib.ui.components.list_view.ListView` and
`cobib.ui.components.search_view.SearchView`.

.. warning::

   This module makes no API stability guarantees! Refer to `cobib.ui.components` for more details.
"""

from __future__ import annotations

from textual.css.query import NoMatches
from textual.widget import Widget
from textual.widgets import ContentSwitcher
from typing_extensions import override


class MainContent(ContentSwitcher):
    """coBib's main content view."""

    DEFAULT_CSS = """
        MainContent {
            width: 2fr;
        }
    """

    # TODO: add a validation hook to the mount method to prevent widgets of "wrong" type?

    def get_current_label(self) -> str:
        """Gets the label of the entry currently under the cursor.

        Raises:
            NoMatches: when `current` is `None`.

        Returns:
            The label of the entry currently under the cursor.
        """
        if self.current is None:
            raise NoMatches
        current_child = self.get_child_by_id(self.current)
        return current_child.get_current_label()  # type: ignore[no-any-return,attr-defined]

    async def replace_widget(self, widget: Widget) -> None:
        """Mounts the provided widget in-place of the one with the same `id`.

        Args:
            widget: the widget to be mounted.
        """
        if widget.id is None:
            return
        old_child = self.get_child_by_id(widget.id)
        await old_child.remove()
        self.mount(widget)
        self.current = widget.id
        self.refresh(layout=True)

    @override
    def notify_style_update(self) -> None:
        super().notify_style_update()
        if self.current is not None:
            current_child = self.get_child_by_id(self.current)
            current_child.notify_style_update()
            current_child.refresh()
